
/*
  Mistral is a constraint satisfaction and optimisation library
  Copyright (C) 2003-2005  Emmanuel Hebrard
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

  The author can be contacted electronically at 
  ehebrard@cse.unsw.edu.au.
*/

#include <sol.h>

/**********************************************
I/   compute length
II/  build the model
III/ create the array 'sequence'
**********************************************/


// prebuild :-> basic initialisation, invoque prebuild on the model

// build :->  n is an upperbound on level max, invoque build 
// on the model and initialise sequence

// initSearch( X ) :-> put the searchable variables in X in front
// of the array variables and returns the number of searched variables.

// initsearch( n, type ) :-> put all variables of the given type
// in front of the array 'variables+n' and the rest at the back.
// return the number of variables whose size is under a given limit


double RPTIMELIMIT=600.0;

/**********************************************
 * Solver
 **********************************************/

void Solver::randomizeSequence()
{
  int i, j, n=(empty-future);
  VariableInt *aux;
  for(i=0; i<n-1; ++i)
    {
      j = i + (rand() % (n-i));
      aux = future[j];
      future[j] = future[i];
      future[i] = aux;
      future[i]->seqIdx = &future[i];
      future[j]->seqIdx = &future[j];
    }

//   for(int k=0; k<length; ++k) {
//     cout << " ";
//     sequence[k]->printshort( cout );
//   }
//   cout << endl;
}

Weighter* Solver::setLearner( int wtype )
  {
    int i=learners.size();
    if( wtype & Weighter::WLD )
      while( i-- )
	if( learners[i]->getType() == Weighter::WDG ) {
	  delete learners[i];
	  learners.erase( i );
	}

    i=learners.size();
    if( wtype & Weighter::WDG )
      while( i-- )
	if( learners[i]->getType() == Weighter::WLD ) {
	  delete learners[i];
	  learners.erase( i );
	}
    
    i=learners.size();
    while( i-- )
      if( learners[i]->getType() & wtype ) 
	return learners[i];
    if( (wtype & Weighter::WLD) == Weighter::WLD ) {
      learners.push( new WeighterLevelDegree( this ) );
    }
    else if( (wtype & Weighter::WDG) == Weighter::WDG ) {
      learners.push( new WeighterDegree( this ) );
    }
    if( wtype & Weighter::IPT) {
      learners.push( new WeighterImpact( this ) );
    }
    if( wtype & Weighter::SAC) {
      learners.push( new WeighterSAC( this ) );
    }
    return learners.back();
  }


void Solver::init()
{
  srand( 11041979 );
  randomizedRestart=true;
  domainSplitting=false;
  verbosity = 0;
  status = UNSAT;
  FIND_ALL = 0;
  SOLUTIONS = NODES = BACKTRACKS = CHECKS = FAILURES = 0; 
  INITTIME = getRunTime();

  BTSLIMIT  =  0;
  NDSLIMIT  =  0;
  TIMELIMIT = .0;
  FAILLIMIT = 0; 

  //  threshold = -1;

#ifdef _MORESTATS

  progressbar = 0;
  LEVELMAX = 0;

#endif

  level      =  0;

  heuristic = NULL;
  goal = NULL;
  //learner = NULL;
}

void Solver::prebuild(CSP& c)
{
  init();
  itsCSP = &c;
  c.preBuild();
}

void Solver::build(CSP& c, const int n)
{
  length = std::min(n, c.numVars());

  c.build( this );
  numvars    =  variables.size();
  sequence = new VariableInt*[numvars];
}


void Solver::initSearch( VarArray& X, const int maxsize ) 
{
  int i=0, j=0, n=X.size(), idx;

  // remove from x the variables that are "non-searchable"
  VariableInt *temp, *aux;
  BuildObject *bvar;
  while( i<n ) {
    bvar = X.array_[i].var_ptr_;
    if( bvar->isSearchable() &&
	((temp = bvar->getVariable()) != NULL) ) {
      
      idx = temp->id;
      if( idx != j ) {
	variables[idx] = variables[j];
	variables[j] = temp;
   
	variables[j]->id = j;
	variables[idx]->id = idx;
      }

      ++j;
    } 
    //else X.array_[i].val = bvar->min();
    ++i;
  }
  n=j;

  initSearch( n, maxsize, false );
}


void Solver::save( ReversibleObj* o )
{
  //store[level].push( o );
  store.push( o );
}

void Solver::initSearch( const int n, 
			 const int maxsize,
			 const bool range ) 
{
  int i, searchable = ( maxsize < INT_MAX ? n : numvars ), end = numvars;

  if( maxsize ) {
    int idx;
    VariableInt* aux;
    while( searchable<end )
      {
	aux = variables[searchable];
	if( aux->isGround() || aux->domsize() > maxsize || 
	    (!range && aux->getType() == VariableInt::RANGE) )
	  {
	    variables[searchable] = variables[--end];
	    variables[searchable]->id = searchable;
	    variables[end] = aux;
	    aux->id = end;
	  }
	else {
	  ++searchable;
	}
      }
  } 

  // then copy the rest of variables into sequence
  std::memcpy(sequence, variables.stack_, numvars*sizeof(VariableInt*));

  empty = (sequence+searchable);
  future = sequence;
  auxilliary = empty;


  length   = (empty-sequence);


#ifdef _MORESTATS

  DOMAINSIZE   = new unsigned int[length];
  for(int i=0; i<length; ++i)
    DOMAINSIZE[i] = sequence[i]->domsize();

  FSTBRANCHING = new unsigned int[length];
  STOBRANCHING = new unsigned int[length];
  AVGBRANCHING = new double[length];

  STONODES = new unsigned int[length];
  LSTNODES = new unsigned int[length];
  std::fill( LSTNODES, LSTNODES+length, 0 );
  AVGNODES = new double[length];
  std::fill( AVGNODES, AVGNODES+length, -1 );

  STOTIME = new double[length];
  AVGTIME = new double[length];

  COUNT = new unsigned int[length];
  std::fill( COUNT, COUNT+length, 0 );

#endif

  gacvarstack.initList(numvars, variables.stack_);
  gacconstack.initStack(constraints.size());


  decision.push( NULL );
  solution = new int[length];
  std::fill( solution, solution+length, INT_MAX );

  lvl_.push(0);
  past.push( future );
  auxv.push( auxilliary );


  for(i=0; i<length; ++i) {
    sequence[i]->seqBeg = &future;
    sequence[i]->seqIdx = sequence+i;
  }

  for(i=0; i<(numvars-length); ++i) {
    auxilliary[i]->seqBeg = &auxilliary;
    auxilliary[i]->seqIdx = auxilliary+i;
  }

  
//   for(i=0; i<constraints.size(); ++i)
//     {
//       VariableInt *w0 = constraints[i]->watch[0];
//       VariableInt *w1 = constraints[i]->watch[1];
//       bool isw0inC = false;
//       bool isw1inC = false;
//       for( int j=0; j<constraints[i]->arity; ++j )
// 	{
// 	  if( constraints[i]->scope[j] == w0 )
// 	    isw0inC = true;
// 	  if( constraints[i]->scope[j] == w1 )
// 	    isw1inC = true;
// 	}

//       assert( isw0inC && isw1inC );

//       if( constraints[i]->watch[0] == constraints[i]->watch[1] ) { //&& !(constraints[i]->watch[0]->isGround()) ) {
//        	constraints[i]->print( cout );
// 	cout << endl;
//       // 	//cout << (constraints[i]->arity) << " " << constraints[i]->id << endl;
//       }
//     }
//   cout << endl;

}
  
Solver::Solver(CSP& c, VarArray& x, VariableOrdering& v) 
{
  prebuild( c );
  build( c, x.size() );
  initSearch( x );
  heuristic = v.extract( this );
}

Solver::Solver(CSP& c, VarArray& x) 
{
  prebuild( c );
  build( c, x.size() );
  initSearch( x );
}

Solver::Solver(CSP& c, VariableOrdering& v) 
{
  prebuild( c );
  build( c );
  initSearch(  );
  heuristic = v.extract( this );
}

Solver::Solver(CSP& c) 
{
  prebuild( c );
  build( c );
  initSearch(  );
}

Solver::Solver() 
{
}

void Solver::add(VariableOrdering& v)
{
  DVO *old_heuristic = heuristic;
  heuristic = v.extract( this );
  if( old_heuristic ) delete old_heuristic;
}


Solver::~Solver() 
{
  delete heuristic;
  delete goal;

  delete [] solution;
  delete [] sequence;
  
  int i;
  for(i=0; i<constraints.size(); ++i) 
    delete constraints[i];
  //  for(i=0; i<unaryCons.size(); ++i) 
  //    delete unaryCons[i];
  for(i=0; i<variables.size(); ++i)
    delete variables[i];


  i=learners.size();
  while( i-- ) 
    {
      delete learners[i];
    }  

//   map<int, VariableInt*>::iterator iter;
//   for( iter = constants.begin(); iter != constants.end(); ++iter)
//     delete iter->second;
}

inline bool Solver::limitsExpired() 
{

#ifdef _DEBUGSEARCH
  if(TIMELIMIT > .0 && getRunTime() - STARTTIME >= TIMELIMIT)
    cout << "TIME LIMIT" << endl;
  
  if(BTSLIMIT  >  0 && BACKTRACKS   >=  BTSLIMIT)    
     cout << "BTS LIMIT" << endl;

  if(NDSLIMIT  >  0 && NODES   >=  NDSLIMIT)    
    cout << "NDS LIMIT" << endl;

  if(FAILLIMIT  >  0 && FAILURES   >=  FAILLIMIT)   
    cout << "FAIL LIMIT" << endl;
#endif

  return (
	  (TIMELIMIT > .0 && getRunTime() - STARTTIME >= TIMELIMIT) ||
	  (BTSLIMIT  >  0 && BACKTRACKS   >=  BTSLIMIT)      ||
	  (NDSLIMIT  >  0 && NODES   >=  NDSLIMIT)      ||
	  (FAILLIMIT  >  0 && FAILURES   >=  FAILLIMIT)      //DG Change
	  );
}


bool Solver::presolve()
{


//   for(int i=0; i<constraints.size(); ++i)
//     {
//       assert( constraints[i]->constrains( constraints[i]->watch[0] ) && 
// 	      constraints[i]->constrains( constraints[i]->watch[1] ) );
//     }
//   cout << endl;
  
  if( !level )
    {
      if( heuristic == NULL )
	heuristic = new DVONoOrder( this );

      STARTTIME = getRunTime();
      
      int i = constraints.size_;
      while( i-- ) {

#ifdef _DEBUGPROPAG
	cout << i << " PROPAGATE \n" ;
	constraints[i]->print( cout );
	cout << endl;
	for(int a=0; a<constraints[i]->arity; ++a) {
	  constraints[i]->scope[a]->print( cout );
	  cout << " ";
	}
	cout << endl;
#endif
	
	++CHECKS;
	if( !(constraints[i]->propagate( )) ) {
	  status = UNSAT;
	  return status;
	}

#ifdef _DEBUGPROPAG
	for(int a=0; a<constraints[i]->arity; ++a) {
	  constraints[i]->scope[a]->print( cout );
	  cout << " ";
	}
	cout << endl;
	constraints[i]->print( cout );
	cout << " OK" << endl;
#endif


      }

      i = unaryCons.size_;
      while( i-- ) {
	++CHECKS;
	if( !(unaryCons[i]->propagate( )) ) {
	  status = UNSAT;
	  return status;
	}
      }
      unaryCons.size_ = 0;
      
      i = numvars;
      while( i-- )
	if( variables[i]->isGround() ) 
	  triggerEvent( variables[i], Constraint::VALUETRIGGER );
	else if( !variables[i]->branch ) {
	  if( domainSplitting && variables[i]->domsize() > 3 && variables[i]->getType() != VariableInt::LIST ) {
	    variables[i]->branch = new ValSelectorSplit( variables[i] );
	  } else {
	    variables[i]->branch = new ValSelectorMin( variables[i] );
	  }
	}
      //   map<int, VariableInt*>::iterator iter;
      //   for( iter = constants.begin(); iter != constants.end(); ++iter)
      //     triggerEvent( iter->second, Constraint::VALUETRIGGER );

      if( !filtering() ) {
	status = UNSAT;
	return false;
      }
      
      ++level;

// #ifdef _MORESTATS
//       cout << "[";
//       for(int i=0; i<pbarlength; ++i)
// 	cout << " ";
//       cout << "]            ";
//       cout.flush();
// #endif

    } 

  return true;
}


int Solver::solve() 
{
  if( presolve() )
    depthFirstSearch( );
  return status;
}


int Solver::ldSolve() 
{
  if( presolve() ) {
    //ConstraintLDS *lds = new ConstraintLDS( this );
    ConstraintLDDS *lds = new ConstraintLDDS( this );
    int ub = lds->getMaxThreshold();

    while( status == UNSAT && ++lds->threshold <= ub ) {

      if( verbosity )
	{
	  TIME = getRunTime() - STARTTIME;
	  cout << "c" << setw(10) << " " << "discrepancy: " 
	       << setw(5) << (lds->threshold) << "  ( " << setw(4) << TIME 
	       << " s)" << endl;
	}

      newNode( NULL );
      ++level;
      
      assert( level == 2 );
    
      depthFirstSearch();
      --level;

      restore();
    }
  }

  return status;
}


int binlog( unsigned long int x ) {
  if( !x ) return INT_MAX;
  if( x == 1 ) return 0;
  int exponent = 0;
  while( (x >> (++exponent)) > 1 );
  //while( ( x >>= 1 ) > 1 )
  //  ++exponent;
  return exponent;
}
unsigned long int luby_seq( unsigned long int iteration ) {
  int thelog = binlog( iteration );
  if( iteration == (1 << (thelog + 1))-1 )
    return (1 << thelog);
  return luby_seq( iteration - (1 << thelog) + 1 );
}

int Solver::solve_and_restart( const int policy, 
			       const unsigned int base, 
			       const double factor)
{

  ////////////////////////////////////////////////////////////
  double PRESOLVETIME = getRunTime();
  ////////////////////////////////////////////////////////////

  if( presolve() ) {

    ////////////////////////////////////////////////////////////
    double PRESOLVETIME = (getRunTime() - PRESOLVETIME);

    //cout << "******* " << PRESOLVETIME << " " << constraints.size() << " ********" << endl;
    if( false && PRESOLVETIME > .3 && PRESOLVETIME/constraints.size() > .0001 ) {
      if( randomizedRestart ) randomizeSequence();
      return depthFirstSearch();
    }
    ////////////////////////////////////////////////////////////
    
    //cout << PRESOLVETIME << endl;
 
    BTSLIMIT = base+BACKTRACKS;
    unsigned long int bts_increment = base, iteration = 0;
    status = LIMITOUT;
    while( status == LIMITOUT ) {
      status = UNSAT;
      ++iteration;
      switch( policy ) {
      case GEOMETRIC : bts_increment = 
	  lround((double)(bts_increment * factor));
	break;
      case LUBY : bts_increment = 
	  lround((double)(base * luby_seq(iteration)));
	break;
      default : bts_increment = 
	  lround((double)(base * luby_seq(iteration) * factor));
      }
      
      
      if( verbosity )
	{
	  TIME = getRunTime() - STARTTIME;
	  cout << "c" << setw(10) << " " << "restart " << setw(12) 
	       << (BTSLIMIT-BACKTRACKS) << "  ( " << setw(4) << TIME 
	       << " s)" << endl;
	}

      //printWeightProfile( cout, 30 );

      //if( randomizedRestart && iteration > 1 ) randomizeSequence();
      if( randomizedRestart ) randomizeSequence();
      depthFirstSearch();

      BTSLIMIT = (BACKTRACKS + bts_increment);
      if( NDSLIMIT > 0 && NODES >= NDSLIMIT || 
	  (TIMELIMIT > .0 && getRunTime() - STARTTIME >= TIMELIMIT) ) break;
    }
  }

  return status;
}


int Solver::random_probe( const unsigned int iterations, 
			  const unsigned int limit )
{
  int rstatus;

  if( presolve() ) {

    int i = iterations;
    status = LIMITOUT;
    
    while( i-- && status == LIMITOUT ) {
      status = UNSAT;

      if( verbosity && !((i+1) % (iterations/10)))
	cout << ".";
      
      FAILLIMIT = FAILURES+limit;

      if( randomizedRestart ) randomizeSequence();      
      depthFirstSearch();

      if( (NDSLIMIT > 0 && NODES >= NDSLIMIT) 
	  || ((getRunTime() - STARTTIME) >= RPTIMELIMIT) 
	  ) 
	break;
    }
    
    FAILLIMIT = 0;
    rstatus = status;
    status = UNSAT;
  }

  return rstatus;
}


int Solver::sacPreprocess( const bool comp, const int wtype )
{

  if( presolve() ) {

    if( wtype & Weighter::WLD == Weighter::WLD )
      setLearner( Weighter::WLD );
    else if( wtype & Weighter::WDG )
      setLearner( Weighter::WDG );
    if( wtype & Weighter::IPT )
      setLearner( Weighter::IPT );

    WeighterSAC *ws = (WeighterSAC*)(setLearner( Weighter::SAC ));

    ws->complete = comp;

    DVOSingletonAC *h = new DVOSingletonAC( this, ws );
    DVO *oldh = heuristic;
    heuristic = h;


    
    ValSelector* oldvo[length];
    for( int i=0; i<length; ++i )
      {
	oldvo[i] = variables[i]->branch;
	if( h->isRange[i] )
	  variables[i]->branch = new ValSelectorSacRange( variables[i], &(h->SACdomain[i]) );
	else
	  variables[i]->branch = new ValSelectorSac( variables[i], &(h->SACdomain[i]) );
      }


    status = LIMITOUT;    
    while( status == LIMITOUT && ws->isNotSac() )
      {
	status = UNSAT;
      
	FAILLIMIT = FAILURES+1;

	depthFirstSearch();
      }
    
    FAILLIMIT = 0;


    delete h;			
    heuristic = NULL;

    if( oldh )
      heuristic = oldh;


    for( int i=0; i<length; ++i )
      {
	delete variables[i]->branch;
	variables[i]->branch = oldvo[i];
      }
    
 //    if( status == LIMITOUT ) {
//       for(int i=0; i<numvars; ++i)
// 	{
// 	  cout << setw(4) << variables[i]->degree 
// 	       << setw(8) << variables[i]->weight ;
	  
// 	  double impact = 0.0;//((DVOSingletonAC*)(s.heuristic))->decision_impact[i][j]; 
// 	  int j = variables[i]->first();
	  
// 	  do {
// 	    impact += wi->decision_impact[i][j]; 
// 	  } while ( variables[i]->setNext(j) );
	  
// 	  cout << setprecision(2) << setw(10) << impact
// 	       << setprecision(2) << setw(10) << (impact / variables[i]->weight) ;
// 	  cout << "   ";
// 	  variables[i]->print( cout );

// 	  j = variables[i]->first();
// 	  do {
// 	    cout << " " << setprecision(2) << setw(5) << wi->decision_impact[i][j] ; 
// 	  } while ( variables[i]->setNext(j) );
	 

// 	  cout << endl;
// 	}
//     }

    delete learners.back();
    learners.pop();
  }


  int rstatus = status;
  status = UNSAT;

  return rstatus;
}


int Solver::checkSolution() 
{
  int i, j, sol[numvars], *aux;
  MistralStack<int> inconsistencies;
  for(i =0; i<constraints.size_; ++i) {
    j = constraints[i]->arity;
    while(j--)
      sol[j] = constraints[i]->scope[j]->first();

    aux = constraints[i]->supp;
    constraints[i]->supp = sol;
    if( !constraints[i]->findSupport(0, sol[0]) ) 
      inconsistencies.push( i );

    constraints[i]->supp = aux;
  }

  while( inconsistencies.size() ) {
    inconsistencies.pop( j );
    cerr << "Error: Solution does not verify\n";
    constraints[j]->print( cerr );
    for(i=0; i<constraints[j]->arity; ++i) {
      cerr << endl;
      cerr << "\t";
      constraints[j]->scope[i]->print( cerr );
    }
    cerr << endl << endl;
    exit( 0 );
  }

  //  ++SOLUTIONS;
  return 1;
}

void Solver::printStatistics(std::ostream &log, const int stats)
{
  TIME = getRunTime() - STARTTIME;
  double TTIME = (TIME + STARTTIME - INITTIME);

  if(TIME < .01)
    TIME = .0;

  if(TTIME < .01)
    TTIME = .0;

  if( stats & OUTCOME ) {
    if( status == LIMITOUT )
      if(FIND_ALL)
	log << std::setw(9) << SOLUTIONS << std::setw(4) << "T_O" ;
      else
	log << std::setw(9) << "TIME_OUT" ;
    else
      if(FIND_ALL)
	log << std::setw(9) << SOLUTIONS << std::setw(4) << "ALL" ;
      else 
	log << std::setw(9) << ((SOLUTIONS > 0) ? ( goal ? "OPT" : "SAT" ) : "UNSAT") ;
  }

  if( stats & BTS )
    log << std::setw(9) << BACKTRACKS << " BTS";
  if( stats & NDS )
    log << std::setw(9) << NODES << " NDS";
  if( stats & FLS )
    log << std::setw(9) << FAILURES << " FAILS";
  if( stats & SPEED )
    log << std::setw(9) << (TIME ? (unsigned int)(((double)BACKTRACKS)/TIME) : (100*BACKTRACKS)) << " BTS/s";
  if( stats & CKS )
    log << std::setw(14) << CHECKS << " CKS";
  if( stats & RUNTIME )
    log << std::setw(8) << std::setprecision(5) << TIME << " s" ;
  if( stats & TOTALTIME )
    log << std::setw(8) << std::setprecision(5) << TTIME << " s" ;
  
  log.flush();
}

void Solver::printWeightProfile(std::ostream& o, int limit) const
{
  int i, j, aux, orderedvar[length];
  for(i=0; i<length; ++i)
    orderedvar[i] = i;
  for(i=1; i<length; ++i)
    {
      j=i;
      while( j && variables[orderedvar[j]]->weight > variables[orderedvar[--j]]->weight )
      //while( j-- && ((double)(variables[orderedvar[j+1]]->domsize())/(double)(variables[orderedvar[j+1]]->weight)) 
      //> ((double)(variables[orderedvar[j]]->domsize())/(double)(variables[orderedvar[j]]->weight)) )
	{
	  aux = orderedvar[j];
	  orderedvar[j] = orderedvar[j+1];
	  orderedvar[j+1] = aux;
	}
    }
  if( length < limit )
    limit = length;
  for(i=0; i<limit; ++i)
    cout << setw(5) << variables[orderedvar[i]]->weight;
  cout << endl;
  for(i=0; i<limit; ++i)
    {
      j = variables[orderedvar[i]]->id;
      if(j < 10)
	cout << "   ";
      else if(j < 100)
	cout << "  ";
      else if(j < 1000)
	cout << " ";
      variables[orderedvar[i]]->printshort( cout );
    }
  cout << endl << endl;
}

bool Solver::preprocess() 
{
  for(int i=0; i<numvars; ++i) 
    gacvarstack.push(variables[i], Constraint::RANGETRIGGER);
  return filtering();
}

int Solver::solutionFound() 
{
  checkSolution();
  ++SOLUTIONS;

  status = SAT;  
  if(goal || verbosity > 1) {

    for(int i=0; i<length; ++i) {
      solution[i] = variables[i]->value() ;
    }

    TIME = getRunTime() - STARTTIME;
    if(TIME < .0001)
      TIME = .0;

    cout << "o " << std::setw(10) << NODES << " NDS"
      //<< std::setw(16) << CHECKS << " CKS"
	 << std::setw(10) << std::setprecision(5)
	 << TIME << " s" ;
 
    cout << "\tsolution(" 
	 << (goal ? goal->score() : SOLUTIONS) << "): {" ;
    for(int i=0; i<length-1; ++i) 
      cout << solution[i] << " ";
    cout << solution[length-1] << "}" << endl;
  }

  if(goal) 
    return (status = goal->update());
  
  if(FIND_ALL && SOLUTIONS != FIND_ALL)
    status = UNSAT;
  else level = 1;

  return status;
}

bool Solver::filtering() 
{

#ifdef _DEBUGSEARCH
  int checksbef = CHECKS;
#endif

  bool consistent = ( !goal || goal->score() <= goal->upper_bound );
  Constraint *con;

  if( consistent ) {
    int sz = unaryCons.size_, domain_event;
    while( consistent && sz-- ) {
      ++CHECKS;
      if( !(unaryCons[sz]->propagate( )) ) consistent = false;
    }

    if( consistent ) {
      VariableInt *pvar;
      MistralNode<Constraint*> *nd;

      bool fixedPoint = (gacvarstack.empty() && gacconstack.empty());     
 
      while( consistent && !fixedPoint ) {

	if( !gacvarstack.empty() ) {

	  pvar = gacvarstack.pop( domain_event );
	  
#ifdef _DEBUGPROPAG
	  if( domain_event & Constraint::VALUETRIGGER ) {
	    cout << "\tValue event on " ;
	    pvar->print( cout );
	    cout << endl;
	    //nd = &(pvar->valueTrigger.head);
	    nd = pvar->constraintsOnValue();
	  } else if( domain_event & Constraint::RANGETRIGGER ) {
	    cout << "\tRange event on " ;
	    pvar->print( cout );
	    cout << endl;
	    //nd = &(pvar->rangeTrigger.head); 
	    nd = pvar->constraintsOnRange(); 
	  } else {
	    cout << "\tDomain event on " ;
	    pvar->print( cout );
	    cout << endl;
	    //nd = &(pvar->domainTrigger.head);
	    nd = pvar->constraintsOnDomain();
	  }
	  cout << endl;
#endif

	  if( domain_event & Constraint::VALUETRIGGER ) {
	    //nd = &(pvar->valueTrigger.head); 
	    nd = pvar->constraintsOnValue();
	    pvar->unLink();
	  } else if( domain_event & Constraint::RANGETRIGGER )
	    //nd = &(pvar->rangeTrigger.head); 
	    nd = pvar->constraintsOnRange(); 
	  else 
	    //nd = &(pvar->domainTrigger.head);
	    nd = pvar->constraintsOnDomain();
	  
	  while( nextNode(nd) ) {
	    con = nd->elt;
	    if( con->delayed ) {
	      gacconstack.push( con, pvar );
	      continue;
	    }

#ifdef _DEBUGPROPAG
	    cout << "PROPAGATE \n" ;
	    con->print( cout );
	    cout << endl;
	    for(int a=0; a<con->arity; ++a) {
	      con->scope[a]->print( cout );
	      cout << " ";
	    }
	    cout << endl;
#endif
	    ++CHECKS;
	    if( !(consistent = con->propagate(pvar, domain_event)) ) {
	      learnFailure( con );
	      break;
	    }

#ifdef _DEBUGPROPAG
	    for(int a=0; a<con->arity; ++a) {
	      con->scope[a]->print( cout );
	      cout << " ";
	    }
	    cout << endl;
	    con->print( cout );
	    cout << " OK" << endl;
#endif
	  }

	  //if( domain_event & Constraint::VALUETRIGGER )
	  // pvar->unLink();
	  
	} else if(consistent) {
	  if( gacconstack.empty() ) fixedPoint = true;
	  else {
	    con = gacconstack.pop( sz );
	    
#ifdef _DEBUGPROPAG 
	    cout << "DELAYED: " ;
	    variables[sz]->print( cout );
	    cout << endl;
	    for(int a=0; a<con->arity; ++a) {
	      con->scope[a]->print( cout );
	      cout << " ";
	    }
	    cout << endl;
	    con->print( cout );
	    cout << endl;
#endif 	
	    ++CHECKS;
	    if( !(consistent = con->propagate(variables[sz], 
					      Constraint::DELAYTRIGGER)) ) {
	      learnFailure( con ); 
	      break;
	    }

#ifdef _DEBUGPROPAG
	    for(int a=0; a<con->arity; ++a) {
	      con->scope[a]->print( cout );
	      cout << " ";
	    }
	    cout << endl;
	    con->print( cout );
	    cout << " OK" << endl;
#endif
	    
	  }
	}
      }
    }
  }

// #ifdef _DEBUGPROPAG
//   if( !consistent ) {
//     if( con )
//       con->print( cout );
//     cout << " FAILS" << endl;
//   }
// #endif

// #ifdef _DEBUGSEARCH
//   for(int k=0; k<(level >> 1); ++k) cout << " ";
//   cout << "GAC: " << (CHECKS-checksbef) << " " 
//        << ( consistent ? "consistent" : "inconsistent" ) << endl;
// #endif


  gacvarstack.clear();
  gacconstack.clear();
  return consistent;
}



int Solver::startNewSearch()
{
  if( !presolve() )
  return status;  
}


int Solver::getNextSolution()
{
  if( future != empty ) {
    status = depthFirstSearch();
  } else {
    if( level == 1 )
      status = UNSAT;
    if( level < decision.size()-1 )
      {

#ifdef _DEBUGSEARCH
	cout << "n";
	for(int k=0; k<level; ++k) cout << " ";
	cout << "descent" << endl;
#endif

	// first descent
	++level;
	if( getNextSolution() == SAT ) return SAT;
	--level;
      }

    restore();

#ifdef _DEBUGSEARCH
    cout << "n";
    for(int k=0; k<level; ++k) cout << " ";
    cout << "restore to level "  << (level-1) << endl;
#endif

    if( limitsExpired() ) { 
      
#ifdef _DEBUGSEARCH
      cout << "n";
      for(int k=0; k<level; ++k) cout << " ";
      cout << "Limit reached" << endl;
#endif    
      
    status = LIMITOUT;
    }
    
    // // Right branch
    if( status != LIMITOUT ) {

#ifdef _DEBUGSEARCH
      cout << "n";
      for(int k=0; k<level; ++k) cout << " ";
      decision[level]->branch->printRight( cout );
      cout << endl;
#endif

      --level;    
      if( decision[level+1]->branch->right() && filtering() ) {
	++level;
	if( depthFirstSearch() == SAT) return SAT;    
      } else {
	++level;
      }
      
      // // Backtrack
      ++BACKTRACKS;
    }

  }
  return status;
}



int Solver::depthFirstSearch() 
{

  // // End of search?
  if( limitsExpired() ) {

#ifdef _DEBUGSEARCH
    cout << "s";
    for(int k=0; k<level; ++k) cout << " ";
    cout << "Limit reached" << endl;
#endif

    return (status = LIMITOUT);
  }
  if( future == empty ) 
    return solutionFound();

  // // Select a variable
  newNode( heuristic->select() );

#ifdef _MORESTATS

  storeMoreStats();

#endif

#ifdef _DEBUGSEARCH
  cout << "s";
  for(int k=0; k<level; ++k) cout << " ";
  decision[level]->branch->printLeft( cout );
  cout << endl;
#endif


  // // Left branch
  decision[level]->branch->left();
  if( filtering() ) {
    learnSuccess();
    ++level;
    if( depthFirstSearch() == SAT ) return SAT;
    --level;
  } 

  restore();

#ifdef _DEBUGSEARCH
  cout << "s";
  for(int k=0; k<level; ++k) cout << " ";
  cout << "restore to level "  << (level-1) << endl;
#endif
  
  if( limitsExpired() ) { 

#ifdef _DEBUGSEARCH
    cout << "s";
    for(int k=0; k<level; ++k) cout << " ";
    cout << "Limit reached" << endl;
#endif    

    status = LIMITOUT;
  }

  // // Right branch
  if( status != LIMITOUT ) {

#ifdef _MORESTATS
    
    updateMoreStats();
    //printProgressBar();

#endif

#ifdef _DEBUGSEARCH
    cout << "s";
    for(int k=0; k<level; ++k) cout << " ";
    decision[level]->branch->printRight( cout );
    cout << endl;
#endif

    --level;    
    if( decision[level+1]->branch->right() && filtering() ) {
      ++level;
      if( depthFirstSearch() == SAT) return SAT;    
    } else {
      ++level;
    }
    
    // // Backtrack
    ++BACKTRACKS;
  }

#ifdef _DEBUGSEARCH
  else {
    cout << "s";
    for(int k=0; k<level; ++k) cout << " ";
    cout << "Limit reached, no right branch" << endl;
  }
#endif

  return status;
}




void Solver::print(std::ostream& o) const
  {
    int i, j;
    o << "Variables: " << endl;
    for(i = 0; i < variables.size(); ++i)
      {
	o << "\t" ;
	variables[i]->print(o);
	o << std::endl;
      }
    o << endl;
    o << "Constraints:" << endl;
    for(i = 0; i < constraints.size(); ++i)
      {
	o << "\t" ;
	constraints[i]->print(o);
	o << std::endl;
      }
    o << endl;
  }

void Solver::printXML(std::ostream& o) const
{
  int i,j;

  // header	  
  o << "<?xml version=\"1.0\" encoding=\"UTF-8\"?>" 
    << endl << endl << "<instance>" << endl 
    << "<presentation name=\"?\" format=\"XCSP 2.0\"/>" 
    << endl << endl;
	
  
  // domains
  int nDomains=0;
  bool known;
  MistralSet domain[length];
  int domainSize[length];
  int domainIndex[length];
  for(i=0; i<length; ++i) 
    {
      known = false;
      for(j=0; !known && j<nDomains; ++j)
	known = ( sequence[i]->included( domain[j] ) && 
		  sequence[i]->include( domain[j] ) );
      if(!known) {	
	domainIndex[i] = nDomains;
	domain[nDomains].init( sequence[i]->min(), 
			       sequence[i]->max(), 
			       MistralSet::empt);
	sequence[i]->unionTo( domain[nDomains] );
	domainSize[i] = sequence[i]->domsize();
	++nDomains;
      } else {
	domainIndex[i] = j;
      }
    }
  o << "<domains nbDomains=\"" << nDomains << "\">" << endl;
  for(i=0; i<nDomains; ++i)
    {
      cout << "<domain name=\"D" << i 
	   << "\" nbValues=\"" << domainSize[i] << "\">" ;
      domain[i].print( cout, "   " );
      cout << "</domain>" << endl; 
    }
  cout << "</domains>" << endl << endl;

}


// int Solver::limitedDiscrepancySearch()
// {
//   // // End of search?
//   if( limitsExpired() ) 
//     return (status = LIMITOUT);
//   if( future == empty ) 
//     return solutionFound();


//   //cout << "left? " << ( gc.member(level) )
//   gc.print( cout );
//   cout << " -> " << (nBadChoices) << " + " << length << " >= " 
//        << threshold << " + " << level << endl;


//   if( !gc.member(level) || (nBadChoices + length >= threshold + level) ) 
//     {

//       // // Select a variable
//       newNode( heuristic->select() );

      
// #ifdef _DEBUGSEARCH
//       cout << "s";
//       for(int k=0; k<level; ++k) cout << " ";
//       decision[level]->branch->printLeft( cout );
//       cout << endl;
// #endif

      
//       // // Left branch
//       decision[level]->branch->left();
//       if( filtering() ) {
// 	learnSuccess();
// 	++level;
// 	if( limitedDiscrepancySearch() == SAT ) return SAT;
// 	--level;
//       } 
      
//       restore();
      
//       if( limitsExpired() ) 
// 	status = LIMITOUT;

//     }


//   // // Right branch
//   if( status != LIMITOUT ) {

//     if( (nBadChoices < threshold) || !gc.member(level) )
//       {

// #ifdef _DEBUGSEARCH
// 	cout << "s";
// 	for(int k=0; k<level; ++k) cout << " ";
// 	decision[level]->branch->printRight( cout );
// 	cout << endl;
// #endif

// 	if( gc.member(level) ) {
// 	  goodChoices.erase(level);
// 	  ++nBadChoices;
// 	}
	
// 	--level;    
// 	if( decision[level+1]->branch->right() && filtering() ) {
// 	  ++level;
// 	  if( limitedDiscrepancySearch() == SAT) return SAT;    
// 	} else {
// 	  ++level;
// 	}

//       }

//     assert( status == UNSAT );

//     // // Backtrack
//     ++BACKTRACKS;
//   }

//   return status;
// }


#ifdef _MORESTATS

double Solver::getProgress()
{
//   long double progress = NODES;
//   for(int i=LEVELMAX-1; i>=0; --i)
//     progress /= AVGBRANCHING[i];




//   long double progress = NODES;
//   int i = 0;
//   while(COUNT[i] < 2)
//     progress /= AVGBRANCHING[i++];
  
//   progress /= (((AVGBRANCHING[i] - (double)(COUNT[i])) * AVGNODES[i]) + NODES);

//   return progress;

  long double tsize = STOBRANCHING[0];
  int i = 0;
  while(COUNT[i] < 2)
    tsize *= STOBRANCHING[++i];
  --tsize;

  if( COUNT[i] >= 2 )
    tsize *= AVGNODES[i];

  tsize += NODES;

  //  cout << ( (double)NODES / tsize ) << endl;

  return ( (double)NODES / tsize );
}

void Solver::storeMoreStats()
{

  if(level > LEVELMAX) LEVELMAX = level;

  ++COUNT[level-1];

  STOBRANCHING[level-1] = decision[level]->domsize();
  AVGBRANCHING[level-1] = ((double)(((COUNT[level-1]-1) * AVGBRANCHING[level-1]) + STOBRANCHING[level-1]) / (double)(COUNT[level-1]));
			   
  STONODES[level-1] = NODES-1;
  STOTIME [level-1] = getRunTime();

}

void Solver::updateMoreStats()
{

  //   cout << "|" << setw(3) << level << "|";  
  //   for(int k=0; k<level; ++k) cout << " ";
  //   if( COUNT[level-1] < 2 )
  //     cout << "-" ;
  //   else
  //     cout << setprecision(2) << AVGNODES[level-1] ;
  
  //   cout << " + " << ((NODES - STONODES[level-1])) ;


  //FSTBRANCHING[level-1] = decision[level]->domsize();;

//   //if( COUNT[level-1] == 2 ) {
//   // this is the 1st time we come back to this level;    
//   STRNODES[level-1] = (NODES - STONODES[level-1]);


//   if( level < LEVELMAX ) {
//     // we are finished with level+1 on this branch;
//     RGTNODES[level] = (RGTNODES[level] * (COUNT[level-1]-1) + STRNODES[level-1] - STRNODES[level]);
//   }

  
  AVGNODES[level-1] = (((double)(COUNT[level-1]-1) * AVGNODES[level-1] + (NODES - STONODES[level-1])) / (double)(COUNT[level-1]));
  
  //  cout << " -> " << setprecision(2) << AVGNODES[level-1] << endl;

  AVGTIME[level-1] = ((double)(COUNT[level-1]-1) * AVGTIME[level-1] + (getRunTime() - STOTIME[level-1])) / (double)(COUNT[level-1]);


  LSTNODES[level-1] = (NODES - STONODES[level-1]);
  

// //   //  long double tsize = 1;
// //   //  for(int k=0; k<LEVELMAX; ++k)
// //   //    tsize *= AVGBRANCHING[k];
// //   //  cout << 227826 << " = " << (unsigned int)tsize << " / " << (227826 - NODES) << endl;


//   cout << setprecision(5) << (getProgress()) << endl;
//   cout << "count:               ";
//   for(int k=0; k<=LEVELMAX; ++k)
//     cout << setw(std::min(8, LEVELMAX+3-k)) << setprecision(2) << COUNT[k];
//   cout << endl;

// //   cout << "fst branching: ";
// //   for(int k=0; k<LEVELMAX; ++k)
// //     cout << setw(std::min(6, LEVELMAX+3-k)) << FSTBRANCHING[k];
// //   cout << endl;

//   cout << "lst branching:       ";
//   for(int k=0; k<LEVELMAX; ++k)
//     cout << setw(std::min(8, LEVELMAX+3-k)) << STOBRANCHING[k];
//   cout << endl;

// // //   cout << "clt branching: ";
// // //   for(int k=0; k<LEVELMAX; ++k) {
// // //     long double tsize = 1;
// // //     for(int p=k; p<LEVELMAX; ++p) {
// // //       tsize *= AVGBRANCHING[p];
// // //       }
// // //     cout << setw(std::min(6, LEVELMAX+3-k)) << (unsigned int)(tsize);
// // //   }
// // //   cout << endl;

// // //   cout << "clt nodes:     ";
// // //   for(int k=0; k<LEVELMAX; ++k)
// // //     cout << setw(std::min(6, LEVELMAX+3-k)) << (LSTNODES[k]);
// // //   cout << endl;

//   cout << "avg  left nodes:     ";
//   for(int k=0; k<LEVELMAX; ++k)
//     {
//       if( AVGNODES[k] > 10 )
//   	{
//   	  if(COUNT[k]>1)
//   	    cout << setw(std::min(8, LEVELMAX+3-k)) << ((int)(AVGNODES[k]));
//   	  else
//   	    cout << setw(std::min(8, LEVELMAX+3-k)) << "-";
//   	}
//       else
//   	{
//   	  if(COUNT[k]>1)
//   	    cout << setw(std::min(8, LEVELMAX+3-k)) << setprecision(2) << AVGNODES[k];
//   	  else
//   	    cout << setw(std::min(8, LEVELMAX+3-k)) << "-";
//   	}
//       }
//   cout << endl;

// //   cout << "avg right nodes:     ";
// //   for(int k=0; k<LEVELMAX-1; ++k)
// //     {
// //       if( (AVGNODES[k] - AVGNODES[k+1]) > 10 )
// //   	{
// //   	  if(COUNT[k]>1)
// //   	    cout << setw(std::min(8, LEVELMAX+3-k)) << ((int)(AVGNODES[k] - AVGNODES[k+1]));
// //   	  else
// //   	    cout << setw(std::min(8, LEVELMAX+3-k)) << "-";
// //   	}
// //       else
// //   	{
// //   	  if(COUNT[k]>1)
// // 	     cout << setw(std::min(8, LEVELMAX+3-k)) << setprecision(2) << (AVGNODES[k] - AVGNODES[k+1]);
// //   	  else
// //   	    cout << setw(std::min(8, LEVELMAX+3-k)) << "-";
// //   	}
// //       }
// //   cout << endl;

//   cout << "lst  left nodes:     ";
//   for(int k=0; k<LEVELMAX; ++k)
//     cout << setw(std::min(8, LEVELMAX+3-k)) << LSTNODES[k];
//   cout << endl << endl;


// //   cout << "lst right nodes:     ";
// //   for(int k=0; k<LEVELMAX-1; ++k)
// //     if(COUNT[k]>1)
// //       cout << setw(std::min(8, LEVELMAX+3-k)) << (int)(LSTNODES[k] - LSTNODES[k+1]);
// //     else
// //       cout << setw(std::min(8, LEVELMAX+3-k)) << "-";
// //   cout << endl << endl;

}

void Solver::printProgressBar()
{
  
  // compute current progress
  double dprog = getProgress();
  unsigned int cprog = (unsigned int)((double)pbarlength * dprog);
  //if( cprog < 0.001 )
  //cprog = 0;

  //  cout <<  progressbar << " " << dprog << " " << ((double)NODES / 227826) << endl;


  if( cprog <= pbarlength && cprog != progressbar ) 
    {
       
      // delete up until progressbar
      for(int i = 13+pbarlength; i > progressbar; --i)
	cout << "\b";
       
      // bad bad bad
      if( cprog < progressbar )
	for(int i = progressbar; i>cprog; --i)
	  cout << "\b";
      
      // print the progress delta
      else 
	for(int i = progressbar; i<cprog; ++i)
	  cout << "=";
      
      // print the blank space
      for(int i=cprog; i<pbarlength; ++i)
	cout << " ";
      cout << "] (";
      double rtime = ((getRunTime() / dprog) * (1-dprog));
      if( rtime < 0.001 )
	rtime = 0;
      if( rtime > 5 ) 
	cout << setw(8) << ((int)rtime) << "s)";
      else
	cout << setprecision(4) << setw(8) << rtime << "s)";


      cout.flush();      

      // update progressbar
      progressbar = cprog;
    
     }
}

// void Solver::print(ostream& o) const
// { 
//   MistralNode<Constraint*> *nd;
//   int i;
//   cout << "Variables: " << endl; 
//   for(i = 0; i < variables.size(); ++i) 
//     { 
//       o << "\t" ; 
//       variables[i]->printshort(o); 
//       o << " " ;
      
//       nd = &(variables[i]->valueTrigger.head);
      
//       while( nextNode(nd) ) {
// 	o << "C" << (nd->elt->id) << " ";
// 	o.flush();
//       }
      
//       o << endl; 
//     } 
// }


#endif

