
#include <sol.h>


/* 
 * a template model
 */

 
using namespace std;

unsigned int nClasses;
unsigned int nExamples;
unsigned int nFeatures;
unsigned int nPositifs;
unsigned int nNegatifs;
unsigned int nNodes;
unsigned int depth;
unsigned int *feature_arity;

map<string,int> classe2index;
map<string,int> feature2index;
map<string,int> *choice2index;


string *index2classe;
string *index2feature;
string **index2choice;

vector< int* > examples;

///////////

int *bin_examples;
unsigned int nBinFeatures;





class ConstraintLearning : public Constraint {

private:
  unsigned int      nExamples; 
  unsigned int      nFeatures; 
  int               *examples;
  ReversibleInt *deepest_node;
  int         **isNotRelevant;
  int                **decide;
  int              *remainder;
  int                coeff[4];
  bool                verbose;


  int isRelevant( const int dir, const int val )
  {
    int negleft=(val << 3) + (dir << 1);
    int negright=negleft+1;
    int posleft=negleft+4;
    int posright=negleft+5;
    return( (remainder[negleft]+remainder[posleft]) && (remainder[negright]+remainder[posright]) );
  }

  int decides( const int dir, const int val )
  {
    int negleft=(val << 3) + (dir << 1);
    int negright=negleft+1;
    int posleft=negleft+4;
    int posright=negleft+5;
    return( !((remainder[negleft]*remainder[posleft]) + 
	      (remainder[negright]*remainder[posright])) );
  }

  
  int getRemainder( const int pol, const int dirup, int dirlo, const int val )
  {
    return remainder[ (val << 3) + (pol << 2) + (dirup << 1) + dirlo ];
  }

  void incrementRemainder( const int pol, const int dirup, const int dirlo, const int val )
  {
    ++remainder[ (val << 3) + (pol << 2) + (dirup << 1) + dirlo ];
  }

  int getCoeff( const int pol, const int dir )
  {
    return coeff[ (pol << 1) + dir ];
  }

  void incrementCoeff( const int pol, const int dir )
  {
    ++coeff[ (pol << 1) + dir ];
  }

//   void setRemainder( const int pol, const int dir, const int val, const int elt )
//   {
//     remainder[ (val << 2) + (pol << 1) + dir ] = elt;
//   }

//   void setRemainder( const int pol, const int dir, const int elt )
//   {
//     remainder[ (nFeatures << 2) + (pol << 1) + dir ] = elt;
//   }

  void clearRemainder()
  {
    std::fill( remainder, remainder+(nFeatures << 3), 0 );
  }

  void clearCoeff()
  {
    std::fill( coeff, coeff+4, 0 );
  }

public:

  static const int left=0;
  static const int right=1;
  static const int pos=1;
  static const int neg=0;

  /**@name Constructors*/
  //@{
  ConstraintLearning(Solver *s,
		     VariableInt **nodes, 
		     const int ln, 
		     int *ex)
    : Constraint(s, nodes, ln, Constraint::VALUETRIGGER)
  {
    verbose = false;
    //delayed = true;
    nExamples = ex[0];
    nFeatures = ex[1];
    examples = &ex[2];
    deepest_node = new ReversibleInt[nExamples];
    isNotRelevant  = new int*[2];
    isNotRelevant[left ]  = new int[nFeatures+1];
    isNotRelevant[right]  = new int[nFeatures+1];
    decide  = new int*[2];
    decide[left ]  = new int[nFeatures];
    decide[right]  = new int[nFeatures];
    for(int i=0; i<nExamples; ++i) {
      deepest_node[i].init( s );
      deepest_node[i].setValue( 0 );

//       cout << setw(2) << i << ":";
//       for(int j=0; j<=nFeatures; ++j)
// 	cout << " " << examples[i*(nFeatures+1)+j];
//       cout << endl;
    }
    //cout << endl;
    remainder = new int[(nFeatures << 3)];
  }
  virtual ~ConstraintLearning()
    {
      delete [] deepest_node;
      delete [] remainder;
    }
  //@}

  void printCoeffs( int node, int feature )
  {
    cout << "current: node" << node << " split on feature" << feature << endl
	 << getCoeff( pos, left ) << "/" << getCoeff( neg, left ) << " <<=  =>> "
	 << getCoeff( pos, right ) << "/" << getCoeff( neg, right ) << endl;
    for(int dir=0; dir<2; ++dir ) {
      if(getCoeff( pos, dir ) && getCoeff( neg, dir ))
	{
	  cout << (dir ? "right: " : "left : ");
	  for(int j=0; j<nFeatures; ++j)
	    {
	      cout << j << ": ";
	      if( getRemainder(pos, dir, left, j) + getRemainder(neg, dir, left, j) )
		{
		  cout << getRemainder( pos, dir, left, j ) << "/" 
		       << getRemainder( neg, dir, left, j ) << " <> ";
		}
	      else
		{
		  cout << " [] <> ";
		}
	      if( getRemainder(pos, dir, right, j) + getRemainder(neg, dir, right, j) )
		{
		  cout << getRemainder( pos, dir, right, j ) << "/" 
		       << getRemainder( neg, dir, right, j ) << " ";
		}
	      else
		{
		  cout << " [] ";
		}

	      assert( isRelevant(dir, j) == !isNotRelevant[dir][j] );

	      if( decides(dir, j) )
		cout << "(d) ";
	      else if( !isRelevant(dir, j) )
		cout << "(r) ";
	      else cout << "    ";
	    }
	  cout << endl;
	}
      else
	{
	  cout << "examples on the " << ( dir ? "right" : "left" ) << " are decided" << endl;
	}
    }
  }


  inline bool propagate(VariableInt *v, const int e) 
  { 
    if( verbose ) {
      cout << "propagate " ;
      v->print( cout ); 
      cout << endl;
    }
    
    bool consistent = true;
    int feature = v->value(), node = v->id, child;
    
    if( feature < nFeatures )
      {
	int i, j;
	
	std::fill(isNotRelevant[right], isNotRelevant[right]+nFeatures+1, 3);
	std::fill(isNotRelevant[left ], isNotRelevant[left ]+nFeatures+1, 3);
	clearRemainder();
	clearCoeff();
	
	int direction, polarity, isDecided[2] = {3,3};	
	
	for(i=0; i<nExamples; ++i) {
	  if(deepest_node[i] == node) {
	    direction = examples[i*(nFeatures+1)+feature];
	    polarity = examples[i*(nFeatures+1)+nFeatures];
	    deepest_node[i] = ((node << 1) + 1 + direction);
	    for(j=0; j<nFeatures; ++j) {
	      isNotRelevant[direction][j] &= (examples[i*(nFeatures+1)+j]+1);
	      incrementRemainder( polarity, direction, examples[i*(nFeatures+1)+j], j );
	    }
	    isDecided[direction] &= (examples[i*(nFeatures+1)+8]+1);
	    incrementCoeff( polarity, direction );
	  }
	}
	  	
	if( verbose ) {
	  printCoeffs( node, feature );
	}

	// look at each branch
	for( direction=0; consistent && direction<2; ++direction )
	  {
	    if( verbose ) {
	      cout << "examine " << (direction ? "right" : "left")  << " branch" << endl;
	    }
	    child = ((node << 1) + 1 + direction);
	    
	    // if not all examples in this subtree are decided 
	    if( consistent && !isDecided[direction] )
	      {		
		// we need at least one level more		
		consistent = ( child < arity );	  

		if( verbose ) {
		  cout << (direction ? "right:" : "left:") 
		       << " need at least node" 
		       << child << endl;
		}

		if( consistent ) {
		  
		  if( verbose ) {
		    cout << "OK" << endl;
		  }

		  // is there a feature that decides all sub-examples?
		  j = nFeatures;
		  while( j-- && !decides(direction, j) );
		  
		  if( j < nFeatures )
		    {

		      if( verbose ) {
			cout << "feature" << j << " decides all axamples " << endl;
		      }

		      // if so set the node to that feature.
		      consistent = scope[child]->setDomain( j );
		    } 
		  else
		    {
		      // the child level is the last, then we fail
		      if( ((child << 1) + 1) >= arity ) 
			{
			  consistent = false;

			  if( verbose ) {
			    cout << "fail because all example should be decided at this level" << endl;
			  }
			}
		      else
			{

			  if( verbose ) {
			    cout << "remove " << nFeatures << " and irrelevant values" << endl;
			  }
			  
			  // otherwise remove the wild card and the irrelevant features
			  consistent = ( scope[child]->remove(nFeatures) );
			  for(j=0; consistent && j<nFeatures; ++j)
			    if( isNotRelevant[direction][j] ) {
			      consistent = scope[child]->remove(j);
			      if( verbose ) {
				cout << "\t" << j << " is not relevant " << endl;
			      }
			    }
			}
		    }
		}
		else {

		  if( verbose ) {
		    cout << "INCONSISTENT" << endl;
		  }
		}
		if( verbose ) {
		  cout << endl;
		}
	      }
	    else
	      {
		consistent = (child >= arity || scope[child]->setDomain(nFeatures) );
	      }
	  }
      } 
    else
      {
	child = ((node << 1) + 1);
	consistent = ( child+1 >= arity || 
		       ( scope[child]->setDomain(nFeatures) && 
			 scope[child+1]->setDomain(nFeatures) ) );
      }

    if( verbose ) {
      for(int i=0; i<depth; ++i)
	{
	  for(int j=(1 << i)-1; j<(1 << (i+1))-1; ++j) {
	    scope[j]->print( cout );
	    cout << " ";
	  }
	  cout << endl;
	}
      cout << "\t=> " << ( consistent ? "OK" : "KO" ) << endl << endl;
    }

     return consistent;
   }
  
  /*!
     Check the corresponding cell in the matrix
  */
  inline int check( const int* s ) const 
  { 
    return 0;
  }

  virtual void print(std::ostream& o) const
  {
    o << "Global Learning Constraint";
  }
  
};


/**********************************************
 * Learning Constraint Declaration
 **********************************************/
class LearningD : public BuildObjectConstraint {

 public:

  virtual void build( Solver *s, VariableInt **tmp, 
		      BuildObjectPredicate *pred )
  {
    Constraint *con = new ConstraintLearning( s, tmp, pred->arity, pred->params );
  }

  virtual void print(std::ostream& o,
		     BuildObject **scope, 
		     const int *params,
		     const int n, const int info) const 
  {
    o << "Global Learning Constraint";
  }

};
/**********************************************
 * Learning Constraint Wrapper
 **********************************************/
class Learning : public Variable {

 public:
 
  Learning( VarArray& x, int *e )
      {
	int n = x.size();
	BuildObject **scope = new BuildObject*[n+1];
	for(int i=0; i<n; ++i) 
	  scope[i] = x[i].var_ptr_;
	var_ptr_ = new BuildObjectPredicate( scope, n, 0, -1, 
					     ENV[ENV.addConstraint( new LearningD() )], e );
					     
      }
  
};




void getData( const char *filename )
{
  string dump;

  int i,j;

  ifstream datafile( filename );
  
  assert( datafile.good() );

  datafile >> dump;
  datafile >> nClasses;

  index2classe = new string[nClasses];
  for(i=0; i<nClasses; ++i)
    {
      datafile >> index2classe[i];
      classe2index[index2classe[i]] = i;
    }

  datafile >> dump;
  datafile >> nFeatures;

  index2feature = new string[nFeatures];
  index2choice = new string*[nFeatures];
  choice2index = new map<string,int>[nFeatures];
  feature_arity = new unsigned int[nFeatures];
  for(i=0; i<nFeatures; ++i)
    {
      datafile >> index2feature[i];
      feature2index[index2feature[i]] = i;
      
      datafile >> feature_arity[i];
      index2choice[i] = new string[feature_arity[i]];
      
      for(j=0; j<feature_arity[i]; ++j)
	{
	  datafile >> index2choice[i][j];
	  choice2index[i][index2choice[i][j]] = j;
	}
    }

  datafile >> dump;
  datafile >> nExamples;
  nPositifs = 0;
  nNegatifs = 0;
  for(i=0; i<nExamples; ++i)
    {
      int *ex = new int[nFeatures+1];
      for(j=0; j<nFeatures; ++j)
	{
	  datafile >> dump;
	  ex[j] = choice2index[j][dump];
	}
      datafile >> dump;
      ex[nFeatures] = classe2index[dump];
      nPositifs += ex[nFeatures];
      nNegatifs += (1-ex[nFeatures]);
      examples.push_back( ex );
    }
}


void getUCIData( const char *filename )
{
  string dump;

  int i,j;

  ifstream datafile( filename );
  
  assert( datafile.good() );

  datafile >> dump;
  datafile >> nClasses;

  index2classe = new string[nClasses];
  for(i=0; i<nClasses; ++i)
    {
      datafile >> index2classe[i];
      classe2index[index2classe[i]] = i;
    }

  datafile >> dump;
  datafile >> nFeatures;

  index2feature = new string[nFeatures];
  index2choice = new string*[nFeatures];
  choice2index = new map<string,int>[nFeatures];
  feature_arity = new unsigned int[nFeatures];
  for(i=0; i<nFeatures; ++i)
    {
      datafile >> index2feature[i];
      feature2index[index2feature[i]] = i;
      
      datafile >> feature_arity[i];
      index2choice[i] = new string[feature_arity[i]];
      
      for(j=0; j<feature_arity[i]; ++j)
	{
	  datafile >> index2choice[i][j];
	  choice2index[i][index2choice[i][j]] = j;
	}
    }

  
  datafile >> dump;
  datafile >> nExamples;
  nPositifs = 0;
  nNegatifs = 0;
  for(i=0; i<nExamples; ++i)
    {
      int *ex = new int[nFeatures+1];
      for(j=0; j<nFeatures; ++j)
	{
	  datafile >> dump;
	  ex[j] = choice2index[j][dump];
	}
      datafile >> dump;
      ex[nFeatures] = classe2index[dump];
      nPositifs += ex[nFeatures];
      nNegatifs += (1-ex[nFeatures]);
      examples.push_back( ex );
    }

}


void printData( )
{
  int i, j;
  cout << "classes: " << nClasses ;
  for(i=0; i<nClasses; ++i)
    cout << " " << index2classe[i];
  cout << endl;

  cout << "features: " << nFeatures << endl;
  
  for(i=0; i<nFeatures; ++i)
    {
      cout << index2feature[i] << " "
	   << feature_arity[i] ;
      for(j=0; j<feature_arity[i]; ++j)
	{
	  cout << " " << index2choice[i][j];
	}
      cout << endl;
    }
  cout << endl;

  cout << "examples: " << nExamples << endl;
  for(i=0; i<nExamples; ++i)
    {
      for(j=0; j<nFeatures; ++j)
	{
	  cout << index2choice[j][examples[i][j]] << " ";
	}
      cout << index2classe[examples[i][nFeatures]] << endl;
    }
  cout << endl;
}


void binaryEncode( )
{
  int i,j,k,l;
  nBinFeatures = 0; 
  for(i=0; i<nFeatures; ++i)
    if(feature_arity[i] > 2)
      nBinFeatures += feature_arity[i];
    else
      ++nBinFeatures;

  bin_examples = new int[(nBinFeatures+1)*examples.size()+2];
  bin_examples[0] = examples.size();
  bin_examples[1] = nBinFeatures;

  int ex[nBinFeatures+1];

  for(i=0; i<examples.size(); ++i)
    {
      k=0;
      for(j=0; j<nFeatures; ++j)
	if(feature_arity[j] > 2) {
	  for(l=0; l<examples[i][j]; ++l)
	    ex[k++] = 0;
	  ex[k++] = 1;
	  for(l=examples[i][j]+1; l<feature_arity[j]; ++l)
	    ex[k++] = 0;
	} else
	  ex[k++] = examples[i][j];
      
      ex[k] = examples[i][nFeatures];

      for(j=0; j<=nBinFeatures; ++j) {
	bin_examples[2+i*(nBinFeatures+1)+j] = ex[j];
      }
    }
}


int main(int argc, const char *argv[])
{  
  int i, j, k, l;
  if( argc <= 1 )
    {
      ofstream outfile( "tmp.lrg", ios_base::out );
      outfile << "classes: 2 + -\nfeatures: 4\noutlook 3 sunny overcast rain\ntemperature 3 hot mild cool\nhumidity 2 high normal\nwindy 2 true false \n\nexamples: 14\novercast hot high false +\nrain mild high false + \nrain cool normal false + \novercast cool normal true + \nsunny cool normal false + \nrain mild normal false + \nsunny mild normal true + \novercast mild high true + \novercast hot normal false + \nrain mild high true - \nsunny hot high false - \nsunny hot high true - \nrain cool normal true - \nsunny mild high false - " << endl;
      getData( "tmp.lrg" );
    }
  else
    getData( argv[1] );

  printData();
  exit( 0 );
  
  binaryEncode( );
  depth = (argc > 2 ? atoi( argv[2] ) : 3 );
  nNodes = (1 << (depth)) - 1;
  
  // create the csp
  CSP model;
  
  VarArray NodeVar( nNodes, 0, nBinFeatures );
  //VarArray ExampleVar( nExamples, 0, nNodes-1 );
  
  if( nPositifs && nNegatifs )
    model.add( NodeVar[0] != nBinFeatures );

  // connect each node in the tree with its children
  // (for the time being with a not-equal)
  for(i=0; i<depth-1; ++i)
    {
      // 	  for(j=(1 << i)-1; j<(1 << (i+1))-1; ++j)
      // 	    cout << j << " ";
      // 	  cout << endl;
      
      for(j=(1 << i)-1; j<(1 << (i+1))-1; ++j)
	{
	  // descendant at level k
	  for(k=1; i+k<depth; ++k)
	    {
	      for(l=0; l<(1 << k); ++l)
		model.add( NotEqualWildcard( NodeVar[j], 
					     NodeVar[((j << k) + (1 << k) -1) + l],
					     nBinFeatures ) );
	    }
	}
    }
  
  model.add( Learning(NodeVar, bin_examples) );
  
  Lexicographic h;
  Solver s( model, NodeVar, h );
  
  if( s.solve() == Solver::SAT )
    {
      int val, width;
      cout << endl;
      for(i=0; i<depth; ++i)
	{
	  
	  width = (int)round((double)(((1 << (depth-1)) * 3) / (double)((1 << i) + 1)));
	  
	  int count = 0;
	  for(j=(1 << i)-1; j<(1 << (i+1))-1; ++j) {
	    ++count;
	    cout << setw(width) ;
	    val = NodeVar[j].value();
	    if( val < nBinFeatures )
	      cout << val ;
	    else
	      cout << " " ;
	  }
	  
	  assert( count == ((1 << i)) );
	  
	  cout << endl;
	}
    }
  
  s.printStatistics( cout );
  cout << endl;
}

