
#include <sol.h>
#include <iostream>
#include <fstream>

using namespace std;

int bufsize = 10000;
int *duration, *machine, nMachines, nJobs, lb, opt;

int maxfsble;
int minfsble;
int makespan;

void readData( char* filename )
{
  int i, j;
  long int dump;
  string tag;
  ifstream infile( filename, ios_base::in );
  
  infile >> nJobs;
  infile >> nMachines;

  duration = new int[nJobs*nMachines];
  machine = new int[nJobs*nMachines];

  infile >> dump;
  infile >> dump;

  infile >> opt ;
  infile >> lb;

  infile >> tag;
  assert( tag == "Times" );

  for(i=0; i<nJobs; ++i) 
    for(j=0; j<nMachines; ++j) 
      infile >> duration[i*nMachines+j];

  infile >> tag;
  assert( tag == "Machines" );

  for(i=0; i<nJobs; ++i) 
    for(j=0; j<nMachines; ++j) {
      infile >> machine[i*nMachines+j];
      --machine[i*nMachines+j];
    }
  
}

 
int upperbound() {
  int i, j, mkp=0, task, boundJob[nJobs], boundMachine[nJobs]; 
  std::fill(boundJob, boundJob+nJobs, 0);
  std::fill(boundMachine, boundMachine+nJobs, 0);


  for(i=0; i<nMachines; ++i)
    {
      for(j=0; j<nJobs; ++j)
	{
	  task = std::max( boundJob[j], boundMachine[machine[j*nMachines+i]] );
	  task += duration[j*nMachines+i];
	  boundJob[j] = task;
	  boundMachine[machine[j*nMachines+i]] = task;
	  if( task > mkp ) mkp = task;
	}
    }
  return mkp;
}


int lowerbound()
{
  int i, j, mkp=0, boundJob[nJobs], boundMachine[nMachines]; 
  std::fill(boundJob, boundJob+nJobs, 0);
  std::fill(boundMachine, boundMachine+nMachines, 0);

  for(i=0; i<nJobs; ++i) {
    for(j=0; j<nMachines; ++j)
      {
	boundJob[i] += duration[i*nMachines+j];
	boundMachine[machine[i*nMachines+j]] += duration[i*nMachines+j];
      }
  }

  for(i=0; i<nJobs; ++i)
    if( mkp < boundJob[i] ) mkp = boundJob[i];
  for(i=0; i<nMachines; ++i)
    if( mkp < boundMachine[i] ) mkp = boundMachine[i];

  return mkp;
}

void setup(CSP& model, VarArray& disjuncts, VarArray& tasks) {
  int i,j,k, per_machine[nJobs*nMachines];

  for(i=0; i<nJobs; ++i) 
    for(j=0; j<nMachines; ++j) 
      tasks.add( Variable(0, makespan-duration[i*nMachines+j]) );
    
  for(i=0; i<nJobs; ++i)
    for(j=0; j<nMachines; ++j) 
      per_machine[i*nMachines+machine[i*nMachines+j]] = (i*nMachines+j);

  for(i=0; i<nJobs; ++i)
    for(j=1; j<nMachines; ++j) 
      model.add( Precedence(tasks[i*nMachines+j-1], duration[i*nMachines+j-1], tasks[i*nMachines+j]) );

  for(i=0; i<nJobs; ++i)
    for(j=0; j<nMachines; ++j) 
      for(k=j+1; k<nMachines; ++k) {
	disjuncts.add( Disjunctive( tasks[per_machine[i*nMachines+j]], 
				    duration[per_machine[i*nMachines+j]],
				    tasks[per_machine[i*nMachines+k]], 
				    duration[per_machine[i*nMachines+k]] ) );
      }

  model.add( disjuncts );
}


int main( int argc, char** argv )
{
  if( argc < 2 )
    {
      cerr << "need a data file" << endl;
      exit( 0 );
    }


  int i, j;
  readData( argv[1] );

  cout << nJobs << " " << nMachines << endl;

  for(i=0; i<nJobs; ++i) {
    for(j=0; j<nMachines; ++j)
      cout << " " << duration[i*nMachines+j];
    cout << endl;
  }

  
  maxfsble = upperbound();
  minfsble = lowerbound();

  cout << minfsble << " <= " << lb << " <= " << opt << " <= " << maxfsble << endl;
  
  // Model

  bool solved = true;
  while( minfsble < maxfsble ) {
    makespan = (int)(floor(((double)minfsble + (double)maxfsble)/2));
      
    cout << "\t" << setw(5) << minfsble << " " << setw(5) << maxfsble 
	 << " (" << setw(5) << makespan << "):\t";
    cout.flush();
                      
    VarArray disjuncts;
    VarArray tasks;
    CSP model;

    setup(model, disjuncts, tasks);

    DomOverWDeg heuristic;
    Solver s(model, disjuncts, heuristic);
    s.setTimeLimit(5);
      

    s.print( cout );
    cout << endl;
    exit( 0 );

    s.setRandomized();
    int result = s.solve_and_restart(Solver::GEOMETRIC, 64, 1.2);
    //int result = s.solve();

    if( result == Solver::SAT ) {
      maxfsble = makespan;
    } else {
      minfsble = makespan+1;
      if( result == Solver::LIMITOUT )
	solved = false;
    }
    
    s.printStatistics(std::cout);
    cout << endl;
  }

  if( !solved ) {
    VarArray disjuncts;
    VarArray tasks;
    VarArray endlasttask( nJobs*nMachines, 0, maxfsble );

    CSP model;
    makespan = maxfsble-1;
    setup(model, disjuncts, tasks);
    for(i=0; i<nJobs; ++i)
      for(j=0; j<nMachines; ++j) 
	model.add( tasks[i*nMachines+j] + duration[i*nMachines+j] == endlasttask[i*nMachines+j] );

    model.add( Minimise(Max(endlasttask)) );

    DomOverWDeg heuristic;
    Solver s(model, disjuncts, heuristic);
    s.solve();
    maxfsble = s.goal->upper_bound;
  }

  cout << maxfsble << " " << opt << endl;


}



  


  




