
#include <sol.h>
//#include <dvo.h>

/*****************************************
 * A model of the NQueens problem
 *****************************************/

/*****************************************
 * Print a solution
 *****************************************/
void printChessboard( VarArray x )
{
  for(int q = 0; q < x.size(); ++q) 
    {
      cout<<"\n\t";
      for(int i = 0; i < x.size(); ++i) 
	{
	  if(x[q].value() == i)
	    cout<<"Q";
	  else
	    cout<<".";
	}   
    }
  cout<<"\n";
}

/*****************************************
 * Model
 *****************************************/
void initChessBoard1( CSP& model, VarArray& queens ) {  
  
  // extra variables  
  VarArray first_diagonal( queens.size() );
  VarArray second_diagonal( queens.size() );
  for(int i=0; i<queens.size(); ++i) {
    first_diagonal[i]  = (queens[i] + i);
    second_diagonal[i] = (queens[i] - i);
  }
  
  // the constraints
  model.add( AllDifferent(first_diagonal) );
  model.add( AllDifferent(second_diagonal) );
  model.add( AllDifferent(queens) );
}

/*****************************************
 * Model
 *****************************************/
void initChessBoard2( CSP& model, VarArray& queens ) {  
  
  // extra variables  
  VarArray first_diagonal( queens.size() );
  VarArray second_diagonal( queens.size() );
  for(int i=0; i<queens.size(); ++i) {
    first_diagonal[i]  = (queens[i] + i);
    second_diagonal[i] = (queens[i] - i);
  }
  
  // the constraints
  for(int i=0; i<queens.size(); ++i) 
    for(int j=i+1; j<queens.size(); ++j) {
      model.add( queens[i] != queens[j] );
      model.add( first_diagonal[i] != first_diagonal[j] );
      model.add( second_diagonal[i] != second_diagonal[j] );
    }
}


// /*****************************************
//  * Model
//  *****************************************/
// void initChessBoard3( CSP& model, VarArray& queens, int t ) {  
  
//   // extra variables  
//   VarArray first_diagonal( queens.size() );
//   VarArray second_diagonal( queens.size() );
//   for(int i=0; i<queens.size(); ++i) {
//     first_diagonal[i]  = (queens[i] + i);
//     second_diagonal[i] = (queens[i] - i);
//   }
  
//   // the constraints
//   model.add( LPAllDifferent(queens, t) );
//   model.add( LPAllDifferent(first_diagonal, t) );
//   model.add( LPAllDifferent(second_diagonal, t) );
// }

/*****************************************
 * Main
 *****************************************/
int main(int argc, char *argv[]) { 
 
  // command line
  int N = ( argc > 1 ? atoi(argv[1]) : 40 );
  int E = ( argc > 2 ? atoi(argv[2]) : 50 );
  int M = ( argc > 3 ? atoi(argv[3]) : 1 );

  //while( N <= E ) {

    // the model
    VarArray queens(N,N);
    CSP model;

    switch( M ) {
    case 1 : { initChessBoard1( model, queens ); break; }
    case 2 : { initChessBoard2( model, queens ); break; }
    default : { initChessBoard1( model, queens ); break; }
    }

    // the solver
    MinDomMin heuristic;

    Solver s( model, queens, heuristic );

//     s.startNewSearch();
//     while( s.getNextSolution() != Solver::UNSAT )
//       {
// 	printChessboard( queens );
// 	cout << endl;
//       }

    s.solve_and_restart(Solver::GEOMETRIC, 32, 2);

    // the solution
    if( N <= 20 )
      printChessboard( queens );
    
    // the statistics
    cout << N << " queens: \t" ;
    s.printStatistics(cout);
    cout << endl ;

    ++N;
    // }
}


