
#include <sol.h>


#ifdef _XERCES

#include "solver/CSPParserCallback.h"
#include "solver/XMLParser.h"
#include "solver/MistralCallback.h"

using namespace CSPXMLParser;

#endif


const int nia = 19;
const char* int_ident[nia] = {
  "-b", 
  "-r", 
  "-rs",

  "-t",
  "-nl",
  
  "-m", 
  "-c",
  "-l", 
  "-s", 

  "-probe", 
  "-pi", 
  "-pl", 
  "-ph",
  "-ulw", 
  "-uip",  

  "-sac", 

  "-fe",
  "-v",
  "-h" 
};

const char* int_man[nia] = {
  "set restart base (default: dynamic)", 
  "randomize the list of variables before each restart (0: no, 1: yes | default: yes)", 
  "set random seed (default: 11041979)\n",

  "set time limit (default: no limit)",
  "set the node limit (default: no limit)\n", 

  "set model type (0: extra variables, 1: no extra variables, 3: dynamic | default: dynamic)", 
  "infers implied AllDifferent constraints (0: no, 1: some, 2: all | default: some)", 
  "set extra variable tolerance (how many extra variables is too many | default: 12000)", 
  "set extra variable tolerance (how big domain is too big | default: 150)\n", 

  "set probing (0: no, 1: yes | default: no)", 
  "set number of probes (default: 100)", 
  "set probe limit (default: 30)", 
  "set heuristic used during probing (0: random, 1: wdeg, 3: wldeg, 8: impact | default: random)", 
  "update level-based weights when probing (0: no, 1: yes | default: no)", 
  "update impacts weights when probing (0: no, 1: yes | default: no)\n", 

  "singleton arc consistency preprocessing (0: no, 1: imcomplete, 2: complete | default: no)\n", 

  "extract features (0: no, 1: yes | default: no)",
  "set verbosity (default: quiet)", 
  "display this help message" 
   };

			      
int int_param[nia];

const int nsa = 4;
const char* str_ident[nsa] = {
  "-name",

  "-heu", 

  "-res",
  "-f" 
};
const char* str_man[nsa] = {
  "set output file extension (default: no output in file)\n",

  "set heuristic (dom, dom/deg, dom/wdeg, dom/wldeg, impact, impact/wdeg, etc.. default: dom/wdeg)\n",
 
  "set restart policy (no, geom, luby, dyn | default: dyn)", 
  "set geometric restart factor (default: 1.333)" 
};
const char* str_param[nsa];


void outputHelpMessage()
{
  cerr << "\nUsage: \t bin/solver [path to the xml file] <args>" << endl << endl;
  for(int i=0; i<nsa; ++i)
    cerr  << str_ident[i] << ":\t" << str_man[i] << endl;
  for(int i=0; i<nia; ++i)
    cerr  << int_ident[i] << ":\t" << int_man[i] << endl;
  cerr << endl;
}

int main(int argc, char **argv)
{

#ifdef _XERCES

  // my interface between the parser and the solver
  MistralCallback cb; 

 try
  {
    // the parser that will decode the XML file
    XMLParser<MistralCallback> parser(&cb); 

    //parser.setPreferredExpressionRepresentation(AST::INFIX_C);
    parser.setPreferredExpressionRepresentation(AST::TREE);
    
    str_param[0] = "";
    getCommandLine(int_ident, int_param, nia,
		   str_ident, str_param, nsa,
		   &(argv[1]), argc-1);


    if( argc < 2 || int_param[18] != INT_MAX || !strcmp( argv[argc-1], "-h" ) )
      {
	outputHelpMessage();
      }
    else
      {
    
	cb.setName         ( strcmp(str_param[0],"nil") ? str_param[0] : "default" );

	cb.setHeuristic    ( strcmp(str_param[1],"nil") ? str_param[1] : "dom/wldeg" );

	cb.setRestartPolicy( strcmp(str_param[2],"nil") ? str_param[2] : "dyn" );
	cb.setRestartFactor( strcmp(str_param[3],"nil") ? atof(str_param[3]) :  (1+1.0/3.0) );
	cb.setRestartBase  ( int_param[0 ] != INT_MAX ? int_param[0 ] : -1 );
	cb.setRandomize    ( int_param[1 ] != INT_MAX ? int_param[1 ] : 1 );
	cb.setRandomSeed   ( int_param[2 ] != INT_MAX ? int_param[2 ] : 11041979 );

	cb.setTimeLimit( (double)(int_param[3] != INT_MAX ? int_param[3] : -1 ) );
	cb.setNodeLimit    ( int_param[4 ] != INT_MAX ? int_param[4] : -1 );

	cb.setModel        ( int_param[5 ] != INT_MAX ? int_param[5 ] :  4 );
	cb.setInferCliques ( int_param[6 ] != INT_MAX ? int_param[6 ] :  1 );
	cb.setEVarLimit    ( int_param[7 ] != INT_MAX ? int_param[7 ] : 12000 );
	cb.setEVarSize     ( int_param[8 ] != INT_MAX ? int_param[8 ] : 150 );

	cb.setProbing      ( int_param[9 ] != INT_MAX ? int_param[9 ] : 0 );
	cb.setPIteration   ( int_param[10] != INT_MAX ? int_param[10] : 100 );
	cb.setPLimit       ( int_param[11] != INT_MAX ? int_param[11] : 30 );
	cb.setProbeHeuris  ( int_param[12] != INT_MAX ? int_param[12] : 0 );
	cb.setUpdateLW     ( int_param[13] != INT_MAX ? int_param[13] : 0 );
	cb.setUpdateIP     ( int_param[14] != INT_MAX ? int_param[14] : 0 );

	cb.setSAC          ( int_param[15] != INT_MAX ? int_param[15] : 0 );

	cb.setFeatureExt   ( int_param[16] != INT_MAX ? int_param[16] : 0 );
	cb.setVerbosity    ( int_param[17] != INT_MAX ? int_param[17] : 0 );
       	
	
	if( argc > 1 )
	  parser.parse( argv[1] ); // parse the input file
	else {
	  cerr << "argument missing" << endl;
	  parser.parse( "/Users/ehebrard/work/dev/mistral/competition/data/newsample/0queens-15.xml" ); 
	}
      }
  }
  catch (exception &e)
  {
    cout.flush();
    cerr << "\n\tUnexpected exception :\n";
    cerr << "\t" << e.what() << endl;
    exit(1);
  }

#endif

  return 0;
}
