#ifndef _MistralCallback_h_
#define _MistralCallback_h_

#include "CSPParserCallback.h"
#include <sol.h>

#define ShowNumericalIDS

namespace CSPXMLParser{
  using namespace std;


  /**********************************************
   * Intentional Constraint
   **********************************************/ 

  class IntentionalConstraint : public Constraint {

  public:
    AST *predicate;
    int *assignment;
    MistralStack<int>* parameters;
    int nbparam;

    /**@name Constructors*/
    //@{
  IntentionalConstraint(Solver *s,
			VariableInt **v, 
			VariableInt **vp, 
			const int l, 
			const int lp,
			AST *p)
    : Constraint(s, v, l)
      {
	nbparam = lp;
	predicate = p;
	assignment = new int[lp];
	std::fill( assignment, assignment+lp, INT_MAX );
	parameters = new MistralStack<int>[arity];
	bool intparam;
	for(int i=0; i<lp; ++i) {
	  intparam = true;
	  for(int j=0; j<arity && intparam; ++j)
	    if( vp[i] == v[j] ) {
	      parameters[j].push( i );
	      intparam = false;
	    }
	  if( intparam ) 
	    assignment[i] = vp[i]->first();
	}
      
	unsigned long int memusage = 1;
	for(int i=0; memusage < 1024 && i<l; ++i)
	  {
	    memusage *= (v[i]->max() - v[i]->min() + 1);
	  }
	if( memusage < 1024 )
	  useResidualSupports(); 
      }
    virtual ~IntentionalConstraint()
      {
	delete [] assignment;
	delete [] parameters;
      }
    //@}


    /*   inline bool propagate(VariableInt *v, const int e) */
    /*   { */
    /*     int i, j; */
    /*     cout << endl;  */
    /*     print( cout ); */
    /*     cout << endl;  */
    /*     for(i=0; i<arity; ++i) */
    /*       for(j=0; j<parameters[i].size_; ++j) { */
    /* 	scope[i]->print( cout ); */
    /* 	cout << " "; */
    /*       } */
    /*     bool res = Constraint::propagate( v, e ); */
    /*     cout << endl;  */
    /*     for(i=0; i<arity; ++i) */
    /*       for(j=0; j<parameters[i].size_; ++j) { */
    /* 	scope[i]->print( cout ); */
    /* 	cout << " "; */
    /*       } */
    /*     cout << endl; */
    /*     print( cout ); */
    /*     cout << " => " << (res ? "OK" : "NI") << endl; */
    /*     cout << endl; */
 
    /*     return res; */
    /*   } */

    /*!
      Check the corresponding cell in the matrix
    */
    inline int check( const int* s ) const 
    { 
      int i, j;
      for(i=0; i<arity; ++i) {
	for(j=0; j<parameters[i].size_; ++j) {
	  assignment[parameters[i][j]] = s[i];
	}
      }
      return !(predicate->value( assignment ));
    }

    virtual void print(std::ostream& o) const
    {
      predicate->infixExpression( o );
      o << "(" ;
      int i,j,k;
      bool isvar;
      for(k=0; k<nbparam; ++k) {
	if( k ) o << ", ";
	isvar = false;
	for(i=0; i<arity; ++i)
	  for(j=0; j<parameters[i].size_; ++j) {
	    if(k == parameters[i][j]) {
	      o << " " ;
	      scope[i]->printshort( o );
	      isvar = true;
	    }
	  }
	if( !isvar )
	  o << assignment[k] ;
      }
      o << " )";
    }
  
  };


  /**********************************************
   * Predicate Constraint Declaration
   **********************************************/
  class PredicateD : public BuildObjectConstraint {

  private:
    int arity;
    AST *tree;
    
  public:

    PredicateD( AST *t, int n ) { arity = n; tree = t;  /* cout << " DECLARE " << n << endl << endl; */ }

    virtual void add( CSP *model, const int l, 
		      BuildObject **scope, 
		      const int n ) 
    {
      //      cout << " ADD " << arity << endl;

      for(int i=arity; i<n; ++i)
	scope[i]->add( model, l+1, (BuildObject::CRANGE | BuildObject::CLIST) );

      //      cout << " ID " << (getVarId()) << endl;
    }
    
    virtual void build( Solver *s, VariableInt **tmp, 
			BuildObjectPredicate *pred )
    {    
      Constraint *con = new IntentionalConstraint( s, tmp, &tmp[arity], arity, 
						   pred->arity-arity, tree ); 
    }
    
    virtual bool  preBuild( BuildObjectPredicate *pred ) 
    {
      bool res = BuildObjectConstraint::preBuild( pred );
      if( res && arity == 2 ) {
	int i, j, k, length = pred->arity-arity;
	int assignment[length];
	int vars[length];
	BuildObject **scope = pred->scope;
	k=0;
	for( i=0; i<length; ++i )
	  if( scope[arity+i] == scope[0] ||
	      scope[arity+i] == scope[1] )
	    vars[k++] = i;
	  else assignment[i] = scope[arity+i]->min();
	
	int borneinf = std::max( pred->scope[0]->min(), pred->scope[1]->min() );
	int bornesup = std::min( pred->scope[0]->max(), pred->scope[1]->max() );
	for( i=borneinf; i<=bornesup; ++i )
	  {
	    for( j=0; j<k; ++j )
	      assignment[vars[j]] = i;
	    if( tree->value( assignment ) ) break;
	  }
	if( i > bornesup ) 
	  pred->setNeq();
      }
      return res;
    }

    virtual void print(std::ostream& o, const BuildObjectPredicate *pred) const

    {
      tree->infixExpression( o );
    }

  };
  /**********************************************
   * Predicate Constraint Wrapper
   **********************************************/
  class Predicate : public Variable {

  public:
 
    Predicate( VarArray& x, VarArray& y, AST *t)
      {
	int n = x.size(), m = y.size();
	BuildObject **scope = new BuildObject*[n+m+1];
	for(int i=0; i<n; ++i) 
	  scope[i] = x[i].var_ptr_;
	for(int i=0; i<m; ++i) 
	  scope[i+n] = y[i].var_ptr_;
	var_ptr_ = new BuildObjectPredicate( scope, n+m, 0, -1, 
					     ENV[ENV.addConstraint( new PredicateD(t,n) )],  
					     NULL, BuildObject::CLIST );//, false, false, false );
	//var_ptr_->setList();
      }
  
  };



  // definition of the functions which are called by the parser when it
  // reads new information in the file. These functions are in charge of
  // creating the data structure that the solver needs to do his job
  class MistralCallback : public CSPParserCallback
  {
  public:
    static const int DYNAMIC = 13;
    
  private:
    
    // timing
    double time_start;
    double time_now;
    double time_last;
    double time_limit;
    int node_limit;

    // parameters/strategies
    unsigned int randomSeed;
    bool   randomized;
    string heuristic;
    int    restart_policy;
    int    restart_base;
    double restart_factor;
    int    model_type;
    int    verbose;
    int    evar_limit;
    int    evar_size;
    int    infer_cliques;

    int    use_sac;
    int    probing;
    int    p_heuristic;
    int    p_iteration;
    int    p_limit;
    bool   updateLW;
    bool   updateIP;
    int    ext_feature;

    string model_name;


    // model
    CSP model;
    VarArray X;

    // constraints
    VarArray constraintScope;
    VarArray predicateScope;
    AST** predicates;
    int consArity;
    int *predicateCount;
    int *predicateScore;

    // network params
    int numVariables;
    int numExtraVars;
    int numDomains;
    int numRelations;
    int numConstraints;
    int numPredicates;
    int numGACPred;
    int numDECPred;
    int numEXTCons;

    // domains
    int **domains;
    int *domainSize;
    int *domainMax;
    int *domainMin;

    // variables;
    int *solution;

    /// old relation stuff
    int *relationSpin;

    /// New relation stuff
    int *relationsIndex;

    // models stats
    int nlistvars;
    int nbitvars;
    int nboolvars;
    int nrangevars;
    int nconstants;

    int elistvars;
    int ebitvars;
    int eboolvars;
    int erangevars;
    int econstants;

    // utilities
    int countI, countJ, countK, maxCount, curCount, idcur;
    CSPDefinitionType curtype;

  public: 

    MistralCallback() {}
    virtual ~MistralCallback() 
      {
      }

    void setRandomize       ( const int rd )   { randomized  = rd; }
    void setRandomSeed      ( const unsigned int rs ) { randomSeed  = rs; }
    void setFeatureExt      ( const int   fe ) { ext_feature = fe; if(fe) verbose = 1; }
    void setModel           ( const int   mt ) { model_type  = mt; }
    void setTimeLimit       ( const double t ) { time_limit  =  t; }
    void setNodeLimit       ( const int    t ) { node_limit  =  t; }
    void setHeuristic       ( const string h ) { heuristic   =  h; }
    void setProbeHeuris     ( const int    h ) { p_heuristic =  h; }
    void setEVarLimit       ( const int    t ) { evar_limit  =  t; }
    void setEVarSize        ( const int    s ) { evar_size   =  s; }
    void setName            ( const string n ) { model_name =  n; }
    void setRestartPolicy   ( string r ) {
      if( r == "no" ) restart_policy = -1;
      else if( r == "geom" ) restart_policy = Solver::GEOMETRIC;
      else if( r == "luby" ) restart_policy = Solver::LUBY;
      else if( r == "dyn" ) restart_policy = DYNAMIC;
      else restart_policy = INT_MAX;
    }
    void setRestartBase  ( int b ) { restart_base = b; }
    void setRestartFactor( double f ) { restart_factor = f;	
      //1+((double)1 / (double)f);
    }
    void setVerbosity    ( int v ) { verbose       = v; }
    void setProbing      ( int p ) { probing       = p; }
    void setPIteration   ( int i ) { p_iteration   = i; }
    void setPLimit       ( int l ) { p_limit       = l; }
    void setUpdateLW     ( int l ) { updateLW      = l; }
    void setUpdateIP     ( int l ) { updateIP      = l; }
    void setSAC          ( int l ) { use_sac       = l; }
    void setInferCliques ( int l ) { infer_cliques = l; }


    virtual void beginInstance(const string & name)
    {
      if( verbose )
	cout << "c \nc Parsing xml file" << endl;
      numExtraVars = numVariables = numDomains 
	= numRelations = numConstraints = numPredicates = 0;

      nlistvars = nbitvars = nboolvars = nrangevars = nconstants =
	elistvars = ebitvars = eboolvars = erangevars = econstants = 
	numGACPred = numDECPred = numEXTCons = 0;

      time_start = getRunTime();

    }

    virtual void beginDomainsSection(int nbDomains)
    {
      if( verbose ) {
	time_now = getRunTime();
	cout << "c" << setw(10) << " " << "domains....." ;
	cout.flush();
      }

      numDomains = nbDomains;
      domains = new int*[numDomains];
      domainSize = new int[numDomains];
      domainMax  = new int[numDomains];
      domainMin  = new int[numDomains];
   
      countI = 0;
      curCount = 0;
      maxCount = nbDomains;
    }
  
    virtual void beginDomain(const string & name,
			     int idDomain,
			     int nbValue)
    {
      domainSize[countI] = nbValue;
      domains[countI] = new int[nbValue];
      countJ = 0;
    }

    void addDomainValue(int v)
    {
      domains[countI][countJ++] = v;
    }

    virtual void addDomainValue(int first,int last)
    {
      for(int k=first; k<=last; ++k)
	domains[countI][countJ++] = k;
    }

    virtual void endDomain()
    {
      domainMax[countI] = -1*INT_MAX;
      domainMin[countI] =    INT_MAX;
      int k=domainSize[countI];
      while( k-- ) {
	if( domains[countI][k] > domainMax[countI] )
	  domainMax[countI] = domains[countI][k];
	if( domains[countI][k] < domainMin[countI] )
	  domainMin[countI] = domains[countI][k];
      }

      ++countI;

      while(curCount < ((int)(10 * countI / maxCount))) {
	++curCount;
	if( verbose ) {
	  cout << ".";
	  cout.flush();
	}
      }
    }

    /*!
     * end the definition of all domains
     */
    virtual void endDomainsSection()
    {
      assert( countI == numDomains );

      if( verbose ) {
	time_now = (getRunTime() - time_now);
	if(time_now < 0.01)
	  time_now = 0;
	cout << setw(6) << setprecision(5)
	     << time_now << "\t(" << numDomains 
	     << " domains)" << endl;
      }
    }


    virtual void beginVariablesSection(int nbVariables)
    {
      if( verbose ) {
	time_now = getRunTime();
	cout << "c" << setw(10) << " " << "variables..." ;
	cout.flush();
      }

      numVariables = nbVariables;
      X.resize( numVariables );
      solution = new int[numVariables];

      countI = 0;
      curCount = 0;
      maxCount = nbVariables;
    }
  
    virtual void addVariable(const string & name, int idVar,
			     const string & domain, int idDomain)
    {

      ++countI;
      solution[idVar] = domainMin[idDomain];

      //cout << countI << "th variable:  in [" << domainMin[idDomain] << ".." << domainMax[idDomain] << "]" << endl;

      X[idVar] = Variable( domains[idDomain], domainSize[idDomain],
			   domainMin[idDomain], domainMax[idDomain] );

      while(curCount < ((int)(10 * countI / maxCount))) {
	++curCount;
	if( verbose ) {
	  cout << ".";
	  cout.flush();
	}
      }
    }
      

    virtual void endVariablesSection()
    {
      assert( countI == numVariables );

      //model.add( X );

      if( verbose ) {
	time_now = (getRunTime() - time_now);
	if(time_now < 0.01)
	  time_now = 0;
	cout << setw(6) << setprecision(5)
	     << time_now << "\t(" << numVariables 
	     << " variables)" << endl;
      }
    }


    virtual void beginRelationsSection(int nbRelations)
    {
      if( verbose ) {
	time_now = getRunTime();
	cout << "c" << setw(10) << " " << "relations..." ;
	cout.flush();
      }

      numRelations = nbRelations;
      relationsIndex = new int[numRelations];
      relationSpin = new int[numRelations];

      countI = 0;
      curCount = 0;
      maxCount = nbRelations;
    }
      
  
    virtual void beginRelation(const string & name, int idRel,
			       int arity, int nbtuples, bool isSupport)
    {
      relationsIndex[countI] = 
	ENV.addConstraint( new BuildObjectTable(arity, nbtuples) );
      relationSpin[countI] = isSupport;
      countJ=0;
    }

    virtual void addRelationTuple(int arity, int tuple[])
    {
      BuildObjectTable *tab = (BuildObjectTable*)(ENV[relationsIndex[countI]]);
      tab->add( tuple );
      ++countJ;
    }

    virtual void endRelation()
    {
      ++countI;

      while(curCount < ((int)(10 * countI / maxCount))) {
	++curCount;
	if( verbose ) {
	  cout << ".";
	  cout.flush();
	}
      }
    }

    virtual void endRelationsSection()
    {
      assert( countI == numRelations );
      
      if( verbose ) {
	time_now = (getRunTime() - time_now);
	if(time_now < 0.01)
	  time_now = 0;
	cout << setw(6) << setprecision(5)
	     << time_now << "\t(" << numRelations 
	     << " relations)" << endl;
      }
    }

    virtual void beginPredicatesSection(int nbPredicates)
    {
      if( verbose ) {
	time_now = getRunTime();
	cout << "c" << setw(10) << " " << "predicates.." ;
	cout.flush();
      }
      
      numPredicates = nbPredicates;
      predicates = new AST*[numPredicates];
      predicateCount = new int[numPredicates];
      predicateScore = new int[numPredicates];
      std::fill(predicateCount, predicateCount+numPredicates, 0);
      std::fill(predicateScore, predicateScore+numPredicates, 0);
      countI = 0;
      curCount = 0;
      maxCount = nbPredicates;
    }
  
    virtual void beginPredicate(const string & name, int idPred)
    {
    }

    virtual void addFormalParameter(int pos, const string & name,
				    const string & type)
    {
    }

    virtual void predicateExpression(AST *tree)
    {
      predicates[countI] = tree;
      tree->eval(predicateCount[countI], 
		 predicateScore[countI]);
    }

    virtual void predicateExpression(const string &expr)
    {
      cerr << "cannot handle expressions in this form" << endl;
      exit(0);
    }

    virtual void endPredicate()
    {
      ++countI;

      while(curCount < ((int)(10 * countI / maxCount))) {
	++curCount;
	if( verbose ) {
	  cout << ".";
	  cout.flush();
	}
      }
    }

    virtual void endPredicatesSection()
    {
      assert( countI == numPredicates );

      if( verbose ) {
	time_now = (getRunTime() - time_now);
	if(time_now < 0.01)
	  time_now = 0;
	cout << setw(6) << setprecision(5)
	     << time_now << "\t(" << numPredicates 
	     << " predicates)" << endl;
      }
    }

    virtual void beginConstraintsSection(int nbConstraints)
    {
      if( verbose ) {
	time_now = getRunTime();
	cout << "c" << setw(10) << " " << "constraints." ;
	cout.flush();
      }

      numConstraints = nbConstraints;
      countI = 0;
      curCount = 0;
      maxCount = nbConstraints;
    }
  
    virtual void beginConstraint(const string & name, int idConstr,
				 int arity,
				 const string & reference,
				 CSPDefinitionType type, int id)
    {
      idcur = id;
      curtype = type;
      countJ=0;
      countK=0;
      consArity = arity;
      constraintScope.clear();
      predicateScope.clear();
    }

    virtual void addVariableToConstraint(const string & name,
					 int idVar)
    {
      constraintScope.add( X[idVar] );
      ++countJ;
    }

    virtual void addEffectiveParameter(int pos, const string & expr)
    {
      cerr << "cannot handle string params" << endl;
      exit(0);
    }

    virtual void addEffectiveParameter(int pos, AST *tree)
    {
      cerr << "cannot handle tree params" << endl;
      exit(0);
    }

    virtual void addEffectiveParameterVariable(int pos,
					       const string & name,
					       int idVar)
    {
      predicateScope.add( X[idVar] );
      ++countK;
    }

    virtual void addEffectiveParameterInteger(int pos,
					      int value)
    {
      predicateScope.add( value );
      ++countK;
    }

    virtual void addEffectiveParameterList(int pos)
    {
      cerr << "cannot handle list params" << endl;
      exit(0);
    }

    /*!
     * indicate the start of a list
     */
    virtual void beginList()
    {
      cerr << "cannot handle list" << endl;
      exit(0);
    }

    virtual void addListVariable(int pos,
				 const string & name,
				 int idVar)
    {
      cerr << "cannot handle list" << endl;
      exit(0);
    }

    virtual void addListInteger(int pos, int value)
    {
      cerr << "cannot handle list" << endl;
      exit(0);
    }

    virtual void endList()
    {
      cerr << "cannot handle list" << endl;
      exit(0);
    }

    virtual void endConstraint()
    {

      if( curtype == RelationType ) {
	++numEXTCons;
	// THOSE ARE THE EXTENSIONAL CONSTRAINTS

	Table con( constraintScope, relationsIndex[idcur], relationSpin[idcur] );
	model.add( con );

  
      } else if( curtype == PredicateType ) { 
	
	/******************************************************
	 * We decide what representation to use amongst:
	 *  + tree of predicate constraints
	 *  + generic AC algorithm using a predicate
	 *  + both methods simultaneously
	 * We use the following criteria:
	 *  - arity
	 *  - size of the cartesian product of the domains
	 *  - size of the predicate tree
	 *******************************************************/

	switch(model_type) {
	case 0 : {
	  ++numDECPred;
	  model.add( Variable( predicates[idcur]->
			       getPredicate( predicateScope ) ) );
	  break;
	}
	case 1 : {
	  ++numGACPred;
	  model.add( Predicate( constraintScope,
				predicateScope,
				predicates[idcur] ) );
	  break;
	}
	case 2 : {
	  ++numGACPred;
	  ++numDECPred;
	  model.add( Variable( predicates[idcur]->
			       getPredicate( predicateScope ) ) );
	  if( consArity < 7 )
	    model.add( Predicate( constraintScope,
				  predicateScope,
				  predicates[idcur] ) );
	  break;
	}
	default : {
	  // helpers
	  int aux, i = consArity;
	  long unsigned cart_cont = 1;

	  /// Attributes:
	  int predArity = predicateScope.size();
	  int rPredArity = predArity;
	  int maxDomSize = 0;
	  long unsigned cartesian_p = 1;
	  int continuity;

	  /// compute cartesian_p, maxDomSize and  continuity
	  while( i-- && cartesian_p <= 10000000 ) {
	    aux = constraintScope[i].var_ptr_->size();
	    if(maxDomSize < aux) maxDomSize = aux;
	    cartesian_p *= aux;
	    cart_cont *= (constraintScope[i].var_ptr_->max() -
			  constraintScope[i].var_ptr_->min() + 1);
	  }
	  continuity = (int)((double)(cart_cont)/(double)(cartesian_p));

	  // compute rPredArity
	  i = predArity;
	  while( i-- )
	    if( predicateScope[i].var_ptr_->size() < 2 )
	      --rPredArity;
	  

	  int predscore = 0;
	  int igacscore = 0;
	  int lastps = 0;
	  int lastgs = 0;


	  if( verbose > 1 ) {
	    predicates[idcur]->infixExpression( cout );
	    cout << endl;
	  }
	  if( verbose > 2 ) {
	    cout << predicateScope << endl;
	    cout << constraintScope << " " << cartesian_p << endl;
	  }

	  // case where we use predicate trees unconditionally
	  if( consArity == 1 || cartesian_p > 10000000 )
	    predscore = INT_MAX;
	  else {


	    /********************************************
	     * Arity
	     *
	     * GAC is better for low arity constraint
	     ********************************************/

	    predscore += (consArity - 2);

	    if( verbose > 2 ) {
	      cout << "Arity:             (" << (predscore - lastps)
		   << " / " << (igacscore - lastgs) << ")" << endl;
	      lastps = predscore;
	      lastgs = igacscore;
	    }


	    /********************************************
	     * "Bad" predicates
	     *
	     * Mul, Div, Mod, Or etc are not very
	     * efficient nor very well implemented,
	     * we should rather use GAC
	     ********************************************/

	    igacscore += ( 3 * (int)((double)(predicateScore[idcur]) / 
				     (double)(predicateCount[idcur])) );

	    if( verbose > 2 ) {
	      cout << "Bad predicates:    (" << (predscore - lastps)
		   << " / " << (igacscore - lastgs) << ")" << endl;
	      lastps = predscore;
	      lastgs = igacscore;
	    }


	    /********************************************
	     * Shape of tree
	     *
	     * Flat trees are better popagated by the decomposition
	     * whilst large trees with a few leaves are better 
	     * propagated using GAC
	     ********************************************/
	    
	    if( predicateCount[idcur] < consArity ) predscore += 1000;
	    igacscore += ( (int)((double)(predicateCount[idcur]) / 
				 (double)(consArity)) );

	    if( verbose > 2 ) {
	      cout << "Shape of the tree: (" << (predscore - lastps)
		   << " / " << (igacscore - lastgs) << ")" << endl;
	      lastps = predscore;
	      lastgs = igacscore;
	    }


	    /********************************************
	     * Continuity
	     *
	     * Domains with holes are often not very well
	     * handled by the decomposition as several
	     * predicates enforces BC instead of GAC.
	     * GAC is much better in this case
	     ********************************************/

	    igacscore += (continuity - 1);

	    if( verbose > 2 ) {
	      cout << "Continuity:        (" << (predscore - lastps)
		   << " / " << (igacscore - lastgs) << ")" << endl;
	      lastps = predscore;
	      lastgs = igacscore;
	    }


	    /********************************************
	     * Search Size
	     *
	     * Large search space, i.e., numerous and/or
	     * big domains may be very costly for a 
	     * generic GAC algorithm. 
	     * The decomposition is better in this case
	     ********************************************/
	    
	    predscore += (int)(log2( cartesian_p ) - 3);
	    
	    if( verbose > 2 ) {
	      cout << "Search Size:       (" << (predscore - lastps)
		   << " / " << (igacscore - lastgs) << ")" << endl;
	      lastps = predscore;
	      lastgs = igacscore;
	    }


	    /********************************************
	     * Domain size
	     *
	     * On Boolean variables the predicates
	     * that enforce BC are equivalent to GAC
	     ********************************************/

	    predscore += 5 * ( maxDomSize <= 2 );
	    
	    if( verbose > 2 ) {
	      cout << "Domain size:       (" << (predscore - lastps)
		   << " / " << (igacscore - lastgs) << ")" << endl;
	      lastps = predscore;
	      lastgs = igacscore;
	    }


	    /********************************************
	     * Number of constraints
	     *
	     * GAC is more memory efficient
	     ********************************************/

	    if( model_type > 3 ) {
	      igacscore += (numConstraints > 32000)*12 + (numConstraints > 64000)*3;
	      
	      if( verbose > 2 ) {
		cout << "Problem size:       (" << (predscore - lastps)
		     << " / " << (igacscore - lastgs) << ")" << endl;
		lastps = predscore;
		lastgs = igacscore;
	      }
	    }


	  }

	    
	  if( predscore > igacscore ) {
	    
	    if( verbose > 1 ) {
	      cout << " ==> DECOMPOSITION " << endl << endl; 
	    }
	    
	    ++numDECPred;
	    model.add( Variable( predicates[idcur]->
				 getPredicate( predicateScope ) ) );
	  } else {
	    
	    if( verbose > 1 ) {
	      cout << " ==> GAC/PREDICATE" << endl << endl; 
	    }

	    ++numGACPred;	    
	    model.add( Predicate( constraintScope,
				  predicateScope,
				  predicates[idcur] ) );
	  }
 	}
	}

      } else if( curtype == GlobalConstraintType ) {
	
	switch( idcur ) {
	case 0 : model.add( AllDifferent(constraintScope) ); 
	  break;
	case 4 : model.add( Element(predicateScope) );
	  break;
	default: {
	  //cerr << endl << constraintScope << endl << predicateScope << endl;
	  

	  cerr << idcur << " not implemented" << endl;
	  exit(0);
	}
	}
      }

      ++countI;
      
      while(curCount < ((int)(10 * countI / maxCount))) {
	++curCount;
	if( verbose ) {
	  cout << ".";
	  cout.flush();
	}
      }
    }

    /**
     * end the definition of all constraints
     */
    virtual void endConstraintsSection()
    {
      if( numRelations ) {
	delete [] relationSpin;
	delete [] relationsIndex;
      }

      for(int k=0; k<numDomains; ++k)
	delete [] domains[k];
      delete [] domains;
      delete [] domainSize;
      delete [] domainMax;
      delete [] domainMin;
      if( numPredicates ) {
	delete [] predicateCount;
	delete [] predicateScore;
      }
      if( verbose ) {
	time_now = (getRunTime() - time_now);
	if(time_now < 0.01)
	  time_now = 0;
	cout << setw(6) << setprecision(5)
	     << time_now << "\t(" << numConstraints 
	     << " constraints)" << endl;
      }
    }

    /********************************************************************/


    void addHeuristic( Solver& s ) {
      if( heuristic == "lex" ) {
	Lexicographic h;
	s.add( h );
      } else if( heuristic == "deg") {
	MaxDegree h;
	s.add( h );
      } else if( heuristic == "rand") {
	Random h;
	s.add( h );
      } else if( heuristic == "dom+deg") {
	MinDomMaxDeg h;
	s.add( h );
      } else if( heuristic == "dom/deg") {
	DomOverDeg h;
	s.add( h );
      } else if( heuristic == "dom/wldeg") {
	DomOverWLDeg h;
	s.add( h );
      } else if( heuristic == "dom/wdeg") {
	DomOverWDeg h;
	s.add( h );
      } else if( heuristic == "neighbor") {
	Neighbor h;
	s.add( h );
      } else if( heuristic == "impact") {
	Impact h;
	s.add( h );
      } else if( heuristic == "impact/wdeg") {
	ImpactOverWDeg h;
	s.add( h );
      } else if( heuristic == "impact/wldeg") {
	ImpactOverWLDeg h;
	s.add( h );
      }
      else {
	MinDomain h;
	s.add( h );
      }
    }


    /*!
     * signal the end of parsing
     */
    virtual void endInstance()
    {


      if( verbose > 1 ) {
	cout << "c\n"
	     << "c" << setw(10) << numEXTCons << " extensional constraints" << endl
	     << "c" << setw(10) << numGACPred << " predicates (GAC)" << endl
	     << "c" << setw(10) << numDECPred << " predicates (decomposition)" << endl;
      }
      if( verbose ) {
	time_now = (getRunTime() - time_start);
	cout << "c Parsing time: " << setw(6) 
	     << setprecision(5) << time_now 
	     << "\nc\nc Building model" << endl;      
      }


      /********************************************
       *
       * Build the model
       *
       ********************************************/

      Solver s;
      s.prebuild( model );

      if( infer_cliques && model.nNeq > 3 && model.nNeq < 8192 ) {
	int nalldiff = model.inferAllDiffs( (infer_cliques > 1), 10000 );
	if( verbose ) 
	  cout << "c" << setw(10) << nalldiff << " AllDifferent constraints infered" << endl;
      }
      
      //cout << evar_limit << " < " << model.numVars() << " - " << X.size() << " / " << model.numEVars() << endl;

      //if( model.numVars() - X.size() > evar_limit ) {
      if( model.numEVars() > evar_limit ) {
	s.build( model, X.size() );
	s.initSearch( X, 0 );
      } else {
	s.build( model );
	s.initSearch( X, evar_size );
      }

      s.INITTIME = time_start;
      s.setTimeLimit( time_limit ); 
      if( node_limit > 0 )
	s.setNodeLimit( node_limit );
      s.setVerbosity( verbose );
      
      if( restart_base <= 0 ) 
	restart_base = std::min(1000, X.size());

      unsigned long int memory_usage;
      unsigned int nvalues = 0;
      unsigned int evalues = 0;

      VariableInt *V;
      if( verbose ) {
	cout << "c" << setw(10) << nconstants << " constants\n";
	cout << "c" << setw(10) << (model.nBoolean) << " Boolean variables    (+" 
	     << (model.eBoolean) << ")\n";
	cout << "c" << setw(10) << (model.nRange)   << " range variables      (+" 
	     << (model.eRange) << ")\n";
	cout << "c" << setw(10) << (model.nBit)     << " domain (b) variables (+" 
	     << (model.eBit) << ")\n";
	cout << "c" << setw(10) << (model.nList)       << " domain (l) variables (+" 
	     << (model.eList) << ")\n";
	cout << "c" << setw(10) << (model.nConstraint) << " constraints          (+" 
	     << (s.constraints.size() - model.nConstraint) << ")\n";
	cout << "c" << setw(10) << nvalues  << " values \n";


#ifdef _LINUX
	memory_usage = getMemory();
#endif

	time_now = (getRunTime() - time_now);
	cout << "c Building time: " << setw(6)
	     << setprecision(5) << time_now 
#ifdef _LINUX
	     << "  (" << (memory_usage >> 8) << " Mb)"  
#endif
	     << "\nc\nc Solving instance" << endl;
	time_now = time_last = getRunTime(); 
      }	  


      if( restart_policy == DYNAMIC )
	{
	  restart_policy = ( (numGACPred + numDECPred) ? Solver::GEOMETRIC : -1 );
	}


      /********************************************
       *
       * Solving
       *
       ********************************************/
      
      if( verbose ) {
	cout << "c" << setw(10) << " " << "heuristic = " << heuristic << endl;
	if( probing ) {
	  cout << "c" << setw(10) << " " << (randomized ? "Randomized " : "" ) 
	       << "probing iteration = " << p_iteration << endl;
	  cout << "c" << setw(10) << " " << "probing limit = " << p_limit << endl;
	  cout << "c" << setw(10) << " " << "level based weights = " << (updateLW ? "yes" : "no") << endl;
	  cout << "c" << setw(10) << " " << "impact probing = " << (updateIP ? "yes" : "no") << endl;
	}
	if( use_sac )
	  cout << "c" << setw(10) << " " << "singleton arc consistency pass " << endl;
	cout << "c" << setw(10) << " " << "restart policy = " ;
	switch( restart_policy ) {
	case -1 : cout << "No restart" << endl; break;
	case Solver::GEOMETRIC : cout << (randomized ? "Randomized / " : "" ) << "Geometric increment" << endl; break;
	case Solver::LUBY : cout << (randomized ? "Randomized / " : "" ) << "Luby policy" << endl; break;
	  //case DYNAMIC : cout << (randomized ? "Randomized " : "" ) << "NO/GEOMETRIC" << endl; break;
	default : cout << "Unknown policy" << endl; break;
	}
	if( restart_policy >= 0 )
	  cout << "c" << setw(10) << " " << "restart base = " << restart_base << endl;
	if( restart_policy == Solver::GEOMETRIC )
	  cout << "c" << setw(10) << " " << "restart factor = " << restart_factor << endl;
      }
      
      s.setRandomSeed( randomSeed );
      if( randomized ) s.setRandomized();
     
	
      int result = Solver::LIMITOUT;
      
      //s.print( cout );
      //cout << endl;
      

      /********************************************
       * Probing for cpzilla
       ********************************************/

      if( ext_feature ) 
	{
	  
	  cout << endl;
	  cout << "c stat Constants:            " << nconstants << endl;
	  cout << "c stat Booleans:             " << nboolvars << endl;
	  cout << "c stat Ranges:               " << nrangevars << endl;
	  cout << "c stat Bitsets:              " << nbitvars << endl;
	  cout << "c stat Lists:                " << nlistvars << endl;
	  cout << "c stat Values:               " << nvalues << endl;
	  cout << "c stat Additional Booleans:  " << eboolvars << endl;
	  cout << "c stat Additional Ranges:    " << erangevars << endl;
	  cout << "c stat Additional Domains:   " << ebitvars << endl;
	  cout << "c stat Additional Values:    " << (model.nValues + model.nConstant - nvalues) << endl;
	  cout << "c stat Constraints:          " << s.constraints.size() << endl;
	  cout << "c stat MaxDepth:             " << s.length << endl;

	  addHeuristic( s );
	  result = s.solve();


	  cout << "c stat Weights:              " ;
	  int i=s.length; 
	  while( i-- )
	    cout << s.variables[i]->weight << " ";
	  cout << endl;


	  cout << "c stat Backtracks:           " << s.BACKTRACKS << endl;
	  cout << "c stat Nodes:                " << s.NODES << endl;
	  cout << "c stat Failures:             " << s.FAILURES << endl;
	  cout << "c stat Checks:               " << s.CHECKS << endl;


#ifdef _MORESTATS

	  cout << "c stat Depth:                " << s.LEVELMAX << endl;
	  cout << "c stat Progress:             " << s.getProgress() << endl;
	  cout << "c stat ExpectedTimeLeft:     " << ((double)time_limit * (1 - (s.getProgress()))) << endl;
	  cout << "c stat AvgBranching:         ";
	  for(int k=0; k<s.LEVELMAX; ++k)
	    cout << setprecision(3) << s.AVGBRANCHING[k] << " ";
	  cout << endl;
	  cout << "c stat AvgNodes:             ";
	  for(int k=0; k<s.LEVELMAX; ++k)
	    if( s.AVGNODES[k] > 10 )
	      cout << (int)(s.AVGNODES[k]) << " ";
	    else
	      cout << setprecision(3) << s.AVGNODES[k] << " ";
	  cout << endl;

#endif

	}
      else 
	{

	  /********************************************
	   * Singleton Arc Consistency
	   ********************************************/
	  
	  if( use_sac )
	    {      
	      if( verbose ) {
		cout << "c" << setw(10) << " " << "singleton arc consistency " ;
		cout.flush();
	      }
	      
	      result = s.sacPreprocess( (use_sac > 1) );
	      
	      if( verbose ) {
		time_now = getRunTime(); 
		cout << setw(6)
		     << setprecision(5) << (time_now - time_last) << endl;
	      }
	    }
	  
	  if( result == Solver::LIMITOUT ) 
	    {
	      /********************************************
	       * Probing
	       ********************************************/
	      if( probing ) 
		{
		  /*********************************************/
		  if( verbose ) {
		    cout << "c" << setw(10) << " " << "probing " ;
		    cout.flush();
		  }
		  /*********************************************/
		  
		  Probe probe_hrst(updateLW, updateIP, p_heuristic);
		  s.add( probe_hrst );
		  result = s.random_probe(p_iteration, p_limit);
		  
		  /*********************************************/
		  if( verbose ) {
		    time_now = getRunTime(); 
		    cout << setw(6)
			 << setprecision(5) << (time_now - time_last) << endl;
		  }
		  /*********************************************/
		}
	      
	      /********************************************
	       * Solving
	       ********************************************/
	      if( result == Solver::LIMITOUT ) 
		{
		  addHeuristic( s );

		  /*********************************************/
		  if( verbose ) {
		    cout << "c" << setw(10) << " " << "final run ";
		    cout.flush();
		  }
		  /*********************************************/

		  if( restart_policy < 0 ) 
		    {
		      result = s.solve();
		    } 
		  else 
		    {
		      if( verbose ) cout << endl;
		      result = s.solve_and_restart(
						   restart_policy,
						   restart_base,
						   restart_factor
						   );
		    }
		  
		  /*********************************************/
		  if( verbose ) {
		    if( restart_policy < 0 ) 
		      {
			time_now = (getRunTime() - time_now); 
			cout << setw(6)
			     << setprecision(5) << time_now << endl;
			cout.flush();
		      }
		  }
		  /*********************************************/
		}
	    }
	  /*********************************************/
	  if( verbose ) {
	    
#ifdef _LINUX
	    memory_usage = (getMemory() - memory_usage);
#endif
	    
	    time_now = getRunTime(); 
	    cout << "c Solving time: " << setw(6)
		 << setprecision(5) << (time_now - time_last) 

#ifdef _LINUX
		 << "  (+" << (memory_usage >> 8) << " Mb)"  
#endif
	      
		 << "\nc\n";
	
	  }
	  /*********************************************/
	}


      /********************************************
       *
       * Print result
       *
       ********************************************/

      bool outputinfile=( model_name != "default" );
      ofstream solfile;
      if( outputinfile) {
	string solname( "sol_" + model_name );
	solfile.open( solname.c_str(), ios_base::out );
      }
      

      int k, finalvalue;
      if( result == Solver::SAT ) {
	if(outputinfile) solfile << "SAT" << endl;
	cout << "SAT" << endl ;
	for(k=0; k<X.size(); ++k) {
	  finalvalue = X[k].value();
	  if( finalvalue != INT_MAX )
	    solution[k] = finalvalue;
	  cout << " " << solution[k];
	  if(outputinfile) solfile << solution[k] << " ";
	}
	if(outputinfile) solfile << endl;
	cout << endl;
      } else if ( result == Solver::UNSAT ) {
	cout << "UNSAT" << endl;
	if(outputinfile) solfile << "UNSAT" << endl;
      } else {
	cout << "UNKNOWN" << endl;
	if(outputinfile) solfile << "UNKNOWN" << endl;
      }
      //cout << "d ";
      s.printStatistics( cout, (Solver::OUTCOME | 
				Solver::TOTALTIME | 
				Solver::BTS |  
				Solver::SPEED |
				Solver::CKS |
				Solver::NDS) );
  
      if(outputinfile) {    
	s.printStatistics( solfile, (Solver::OUTCOME |
				     Solver::TOTALTIME |
				     Solver::BTS |
				     Solver::CKS |
				     Solver::NDS) );
	solfile << endl;
      }
      cout << endl;
      
      if( numPredicates ) {
	for(int i=0; i<numPredicates; ++i)
	  delete predicates[i];
	delete [] predicates;
      }
      delete [] solution;
    }

  };

} // namespace

#endif

